import ase
import numpy as np
import pytest
from aiida import orm

from aiida_vasp.utils.kmesh import get_ir_kpoints_and_weights, get_ir_kpoints_data


@pytest.fixture
def atoms():
    """A test structure"""
    lattice = np.array([[0.0, 0.5, 0.5], [0.5, 0.0, 0.5], [0.5, 0.5, 0.0]]) * 5.4
    positions = [[0.875, 0.875, 0.875], [0.125, 0.125, 0.125]]
    numbers = [
        1,
    ] * 2
    return ase.Atoms(cell=lattice, scaled_positions=positions, numbers=numbers)


@pytest.fixture
def structure(atoms):
    return orm.StructureData(ase=atoms)


def test_get_ir_kpoints_from_structure(atoms):
    """Test the get_ir_kpoints_from_structure function"""

    mesh = [8, 8, 8]

    # Test gamma-center grids
    ir_kpoints, weights = get_ir_kpoints_and_weights(atoms.cell, atoms.get_scaled_positions(), atoms.numbers, mesh)
    assert ir_kpoints.shape == (29, 3)
    assert np.sum(weights) == 1
    # Test shifted grids
    ir_kpoints, weights = get_ir_kpoints_and_weights(
        atoms.cell, atoms.get_scaled_positions(), atoms.numbers, mesh, is_shift=[1, 1, 1]
    )
    assert ir_kpoints.shape == (60, 3)
    assert np.sum(weights) == 1

    # Test not doing reduction at all
    ir_kpoints, weights = get_ir_kpoints_and_weights(
        atoms.cell, atoms.get_scaled_positions(), atoms.numbers, mesh, symmetry_reduce=False
    )
    assert ir_kpoints.shape == (8 * 8 * 8, 3)
    assert np.sum(weights) == 1


def test_get_ir_kpoints_data(structure):
    """Test the get_ir_kpoints_data function"""
    mesh = [8, 8, 8]
    kpt = get_ir_kpoints_data(structure, mesh)
    coords, weights = kpt.get_kpoints(also_weights=True)
    assert coords.shape == (29, 3)
    assert sum(weights) == 1

    kpt = get_ir_kpoints_data(structure, mesh, is_shift=[1, 1, 1])
    coords, weights = kpt.get_kpoints(also_weights=True)
    assert coords.shape == (60, 3)
    assert sum(weights) == 1

    kpt = get_ir_kpoints_data(structure, mesh, symmetry_reduce=False)
    coords, weights = kpt.get_kpoints(also_weights=True)
    assert coords.shape == (8 * 8 * 8, 3)
    assert sum(weights) == 1



"""Test the POTCAR parser."""

import pytest

from aiida_vasp.data.potcar import PotcarData, PotcarFileData
from aiida_vasp.parsers.content_parsers.potcar import MultiPotcarIo, PotcarIo


def verify_potcario(potcario):
    assert potcario.node
    assert potcario.file_node
    assert potcario.content


def test_potcar_from_path(aiida_profile, data_path):
    """Create a PotcarIo instance from a file path."""
    potcar_path_as = data_path('potcar', 'As', 'POTCAR')
    from_ctor = PotcarIo(path=potcar_path_as)
    from_from = PotcarIo.from_(potcar_path_as)
    verify_potcario(from_from)
    assert from_ctor == from_from


def test_potcar_from_file_node(upload_potcar):
    """Create a PotcarIo instance from a PotcarFileData node."""
    potcar_file_in = PotcarFileData.find_one(element='In')
    from_ctor = PotcarIo(potcar_file_node=potcar_file_in)
    verify_potcario(from_ctor)
    from_from = PotcarIo.from_(potcar_file_in)
    assert from_ctor == from_from


def test_potcar_from_node(upload_potcar):
    """Create a PotcarIo instance from a PotcarData node."""
    potcar_ga = PotcarData.find_one(element='Ga')
    from_ctor = PotcarIo(potcar_node=potcar_ga)
    verify_potcario(from_ctor)
    from_from = PotcarIo.from_(potcar_ga)
    assert from_ctor == from_from


def test_potcar_from_contents(upload_potcar, read_file):
    """Create a PotcarIo from contents of a POTCAR file."""
    contents_as = read_file('potcar', 'As', 'POTCAR')
    from_ctor = PotcarIo(contents=contents_as.encode('utf-8'))
    verify_potcario(from_ctor)
    assert from_ctor.node.uuid == PotcarData.find_one(element='As').uuid
    from_from = PotcarIo.from_(contents_as)
    assert from_ctor == from_from


def test_file_contents_equivalence(aiida_profile, data_path, read_file):
    potcar_path_as = ['potcar', 'As', 'POTCAR']
    from_file = PotcarIo(path=data_path(*potcar_path_as))
    from_contents = PotcarIo(contents=read_file(*potcar_path_as).encode('utf-8'))
    assert from_file.sha512 == from_contents.sha512


def test_multi_round_trip(upload_potcar, potcar_family_name, tmp_path, potcar_mapping):
    """Write multiple POTCAR potentials to a file and recover the nodes stored in the db."""
    test_dir = tmp_path / 'round_trip'
    test_dir.mkdir()
    potcar_cls = PotcarData
    multi = MultiPotcarIo(
        potcar_cls.get_potcars_dict(
            elements=potcar_mapping.keys(), family_name=potcar_family_name, mapping=potcar_mapping
        ).values()
    )
    tempfile = test_dir / 'POTCAR'
    multi.write(tempfile)
    recovered = multi.read(tempfile)
    uuids_start = [potcar.node.uuid for potcar in multi.potcars]
    uuids_recov = [potcar.node.uuid for potcar in recovered.potcars]
    assert uuids_start == uuids_recov


@pytest.mark.parametrize(['vasp_structure'], [('str',)], indirect=True)
def test_multi_from_structure(upload_potcar, potcar_family_name, vasp_structure_poscar, potcar_mapping):
    potcar_cls = PotcarData
    potcar_dict = potcar_cls.get_potcars_dict(
        elements=['As', 'In', 'In_d'], family_name=potcar_family_name, mapping=potcar_mapping
    )
    multi = MultiPotcarIo.from_structure(structure=vasp_structure_poscar._content_data, potentials_dict=potcar_dict)  # pylint: disable=protected-access
    assert [potcar.node.full_name for potcar in multi.potcars] == ['In_sv', 'As', 'In_d', 'As']



"""
Tests for mock vasp
"""

# pylint: disable=unused-import,redefined-outer-name,unused-argument,unused-wildcard-import,wildcard-import,no-member, import-outside-toplevel, too-many-arguments
import pathlib
from shutil import copy2, rmtree
from tempfile import mkdtemp

import pytest

from aiida_vasp.utils.mock_code import MockVasp, VaspMockRegistry, get_hash


def test_get_hash():
    """
    Test the get_hash function
    """
    dict1 = {'1': 2, 3: 4, 'a': [1, '2', '3']}
    hash1 = get_hash(dict1)[0]
    dict2 = {'1': 38, 3: 4, 'a': [1, '2', '3']}
    hash2 = get_hash(dict2)[0]
    dict3 = {3: 4, '1': 2, 'a': [1, '2', '3']}
    hash3 = get_hash(dict3)[0]

    assert hash1 != hash2
    assert hash1 == hash3


def test_get_hash_list():
    """Test generating has for nested dict/list"""
    dict1 = {'1': 2, 3: 4, 'a': [1, -0.0, '3']}
    hash1 = get_hash(dict1)[0]
    dict2 = {'1': 2, 3: 4, 'a': [1, 0.0, '3']}
    hash2 = get_hash(dict2)[0]

    assert hash1 == hash2

    dict1 = {'1': 2, 3: 4, 'a': [1, -0.0, '3', {'1': 0.0}]}
    hash1 = get_hash(dict1)[0]
    dict2 = {'1': 2, 3: 4, 'a': [1, 0.0, '3', {'1': -0.0}]}
    hash2 = get_hash(dict2)[0]

    assert hash1 == hash2


@pytest.fixture(scope='module')
def mock_registry():
    """
    Get an mock registry object
    """
    return VaspMockRegistry()


@pytest.fixture
def custom_registry():
    """
    Return an temporary registry
    """
    temp_base = mkdtemp()
    yield VaspMockRegistry(base_path=pathlib.Path(temp_base))
    rmtree(temp_base)


@pytest.fixture
def temp_path():
    """Return an temporary folder"""
    temp_base = mkdtemp()
    yield pathlib.Path(temp_base)
    rmtree(temp_base)


def test_registry_scan(mock_registry):
    """
    Test repository scanning
    """
    mock_registry.scan()
    assert len(mock_registry.reg_hash) > 0
    # Check some existing mocks are there already
    assert 'test_bands_wc' in mock_registry.reg_name


def test_registry_extract(mock_registry):
    """Test extracting an folder from the registry"""

    tmpfolder = mkdtemp()
    mock_registry.extract_calc_by_path('test_bands_wc', tmpfolder)
    objects = [path.name for path in pathlib.Path(tmpfolder).glob('*')]
    assert 'OUTCAR' in objects
    assert 'vasprun.xml' in objects
    assert 'INCAR' in objects

    rmtree(tmpfolder)


def test_registry_match(mock_registry):
    """Test round-trip hash compute and matching"""

    hash_val = mock_registry.compute_hash(mock_registry.base_path / 'test_bands_wc/inp')
    assert hash_val in mock_registry.reg_hash


def test_registry_folder_upload(mock_registry, custom_registry, temp_path):
    """Test uploading a folder to the registry"""

    # Exact an existing calculation to the folder
    mock_registry.extract_calc_by_path('test_bands_wc', temp_path)
    # Upload to a different registry
    custom_registry.upload_calc(temp_path, 'upload-example')

    # Reset the direcotry
    rmtree(str(temp_path))
    temp_path.mkdir()

    # Extract and validate
    assert 'upload-example' in custom_registry.reg_name
    custom_registry.extract_calc_by_path('upload-example', temp_path)
    objects = [path.name for path in temp_path.glob('*')]

    assert 'OUTCAR' in objects
    assert 'vasprun.xml' in objects
    assert 'INCAR' in objects


@pytest.mark.parametrize(['vasp_structure', 'vasp_kpoints'], [('str', 'mesh')], indirect=True)
def test_registry_upload_aiida(run_vasp_process, custom_registry, temp_path):
    """Test upload from an aiida calculation"""

    _, node = run_vasp_process()
    custom_registry.upload_aiida_calc(node, 'upload-example')

    # Exact the calculation
    custom_registry.extract_calc_by_path('upload-example', temp_path)

    objects = [path.name for path in temp_path.glob('*')]
    assert 'OUTCAR' in objects
    assert 'vasprun.xml' in objects
    assert 'INCAR' in objects


@pytest.mark.parametrize(['vasp_structure', 'vasp_kpoints'], [('str', 'mesh')], indirect=True)
def test_registry_upload_wc(run_vasp_process, custom_registry, temp_path):
    """Return an VaspWorkChain node that has been executed."""
    _, node = run_vasp_process(process_type='workchain')
    custom_registry.upload_aiida_work(node, 'upload-example')
    # Extract the calculation
    repo_path = custom_registry.get_path_by_name('upload-example/calc-000')
    objects = [path.name for path in repo_path.glob('out/*')]
    assert 'OUTCAR' in objects
    assert 'vasprun.xml' in objects

    custom_registry.extract_calc_by_path('upload-example/calc-000', temp_path)
    objects = [path.name for path in temp_path.glob('*')]
    assert 'OUTCAR' in objects
    assert 'vasprun.xml' in objects
    assert 'INCAR' in objects


def test_mock_vasp(mock_registry, temp_path, data_path):
    """Test the MockVasp class"""

    # Setup the input directory
    mock_vasp = MockVasp(temp_path, mock_registry)
    base_path = pathlib.Path(data_path('test_bands_wc', 'inp'))

    with pytest.raises(ValueError):
        # Should fail due to not having any input files present and
        # we can then not match it to a test dataset
        mock_vasp.run()

    for obj in ['INCAR']:
        copy2(base_path / obj, temp_path / obj)

    with pytest.raises(ValueError):
        # Should fail due to missing POSCAR
        mock_vasp.run()

    for obj in ['POSCAR']:
        copy2(base_path / obj, temp_path / obj)

    mock_vasp.run()

    objects = [path.name for path in temp_path.glob('*')]
    assert 'OUTCAR' in objects
    assert 'vasprun.xml' in objects



"""
Common click options for verdi.

Common click options for verdi commands.
"""

from __future__ import annotations

from typing import Any

import click

try:
    from aiida.cmdline.params.options import DESCRIPTION, FORCE, OverridableOption  # pylint: disable=unused-import
except ImportError:
    # pylint: disable=too-few-public-methods
    class OverridableOption:  # pylint: disable=useless-object-inheritance
        """
        Wrapper around click option that increases reusability.

        Click options are reusable already but sometimes it can improve the user interface to for example customize a
        help message for an option on a per-command basis. Sometimes the option should be prompted for if it is not
        given. On some commands an option might take any folder path, while on another the path only has to exist.

        Overridable options store the arguments to click.option and only instantiate the click.Option on call, kwargs
        given to ``__call__`` override the stored ones.

        Example::

            FOLDER = OverridableOption('--folder', type=click.Path(file_okay=False), help='A folder')

            @click.command()
            @FOLDER(help='A folder, will be created if it does not exist')
            def ls_or_create(folder):
                click.echo(os.listdir(folder))

            @click.command()
            @FOLDER(help='An existing folder', type=click.Path(exists=True, file_okay=False, readable=True)
            def ls(folder)
                click.echo(os.listdir(folder))
        """

        def __init__(self, *args: Any, **kwargs: Any) -> None:
            """Store the defaults."""
            self.args = args
            self.kwargs = kwargs

        def __call__(self, **kwargs: Any) -> click.Option:
            """Override kwargs (no name changes) and return option."""
            kw_copy = self.kwargs.copy()
            kw_copy.update(kwargs)
            return click.option(*self.args, **kw_copy)

    FORCE = OverridableOption('-f', '--force', is_flag=True, help='Do not ask for confirmation.')
    DESCRIPTION = OverridableOption('-d', '--description', help='(text) description')

FAMILY_NAME = OverridableOption(
    '-n',
    '--name',
    required=True,
    help='The name you want to give this family of potential(s). '
    'This is the name you will use in the future when telling which potentials to use '
    'for a specific calculation or workflow.',
)
PATH = OverridableOption('-p', '--path', default='.', type=click.Path(exists=True), help='Path to the folder.')

DRY_RUN = OverridableOption(
    '--dry-run', is_flag=True, is_eager=True, help='Do not commit to database or modify configurations.'
)



from __future__ import annotations

import os
import pathlib
import subprocess as sp

import numpy as np
import pytest
from aiida import orm
from aiida.cmdline.utils.ascii_vis import format_call_graph
from aiida.cmdline.utils.common import get_calcjob_report, get_node_info, get_workchain_report
from aiida.common.exceptions import NotExistent
from aiida.common.extendeddicts import AttributeDict
from aiida.engine import run
from aiida.orm import CalculationNode, Code, Computer, Dict, InstalledCode, QueryBuilder, load_code
from aiida.plugins import CalculationFactory, DataFactory, WorkflowFactory
from aiida.tools.archive import create_archive

from aiida_vasp.data.potcar import OLD_POTCAR_FAMILY_TYPE, Group, PotcarData, PotcarFileData, PotcarGroup, temp_potcar
from aiida_vasp.utils.general import copytree
from aiida_vasp.workchains.v2.common.builder_updater import VaspBuilderUpdater

pytest_plugins = 'aiida.tools.pytest_fixtures'

cwd = pathlib.Path(__file__).parent


@pytest.fixture()
def fresh_aiida_env(aiida_profile_clean):
    """Reset the database before and after the test function."""
    try:
        yield aiida_profile_clean
    finally:
        print_and_export_failed_mock()


@pytest.fixture(scope='session')
def localhost_dir(tmp_path_factory):
    return tmp_path_factory.mktemp('localhost_work')


@pytest.fixture()
def localhost(aiida_profile, localhost_dir):
    """Fixture for a local computer called localhost. This is currently not in the AiiDA fixtures."""
    try:
        computer = Computer.collection.get(label='localhost')
    except NotExistent:
        computer = Computer(
            label='localhost',
            hostname='localhost',
            transport_type='core.local',
            scheduler_type='core.direct',
            workdir=str(localhost_dir),
        ).store()
        computer.set_minimum_job_poll_interval(0.0)
        computer.configure()
    return computer


@pytest.fixture()
def mock_vasp(aiida_profile, localhost):
    """
    Give an mock-up of the VASP executable

    This code will always create the output object even if no matching
    calculations from the registry is found. This makes it suitable for simple
    tests.
    """
    return _create_mock_vasp_code(aiida_profile, localhost, 'mock-vasp-loose')


@pytest.fixture()
def mock_vasp_strict(aiida_profile, localhost):
    """
    Give an mock-up of the VASP executable with strict input matching.

    This code will not create the output object unless matching calculations from the
    registry is found. It is suitable for testsing complex multi-step workchains.
    tests.
    """
    return _create_mock_vasp_code(aiida_profile, localhost, 'mock-vasp')


@pytest.fixture()
def data_path():
    """Give the path to test data."""

    def _data_path(*args):
        path = pathlib.Path(cwd / 'test_data', *args)
        path = path.resolve()
        assert path.exists()
        assert path.is_absolute()
        return str(path)

    return _data_path


@pytest.fixture()
def read_file(data_path):
    """Give the content (string) of a test data file."""

    def _read_file(*args, **kwargs):
        path = kwargs.get('path', None)
        mode = kwargs.pop('mode', None)
        encoding = kwargs.pop('encoding', None)
        if not mode:
            mode = 'r'
        if not path:
            path = data_path(*args)
        if encoding is not None and 'b' not in mode:
            encoding = 'utf8'
        with open(path, mode, encoding=encoding) as testdata_fo:
            testdata_content = testdata_fo.read()
        return testdata_content

    return _read_file


@pytest.fixture()
def vasp_code(localhost):
    """Fixture for a vasp code, the executable it points to does not exist."""

    if not localhost.pk:
        localhost.store()
    code = InstalledCode(localhost, '/usr/local/bin/vasp')
    code.label = 'vasp'
    code.description = 'VASP code'
    code.default_calc_job_plugin = 'vasp.vasp'
    return code


@pytest.fixture
def vasp_params(aiida_profile):
    incar_data = orm.Dict(dict={'gga': 'PE', 'gga_compat': False, 'lorbit': 11, 'sigma': 0.5, 'magmom': '30 * 2*0.'})
    return incar_data


@pytest.fixture(params=['cif', 'str'])
def vasp_structure(request, aiida_profile, data_path):
    """Fixture: StructureData or CifData."""
    if request.param == 'cif':
        cif_path = data_path('cif', 'EntryWithCollCode43360.cif')
        structure = DataFactory('core.cif').get_or_create(cif_path)[0]
    elif request.param == 'str':
        larray = np.array([[0, 0.5, 0.5], [0.5, 0, 0.5], [0.5, 0.5, 0]])
        alat = 6.058
        structure = DataFactory('core.structure')(cell=larray * alat)
        structure.append_atom(position=[0, 0, 0], symbols='In')
        structure.append_atom(position=[0.25, 0.25, 0.25], symbols='As')
        structure.append_atom(position=[0.25, 0.33, 0.34], symbols='As')
        structure.append_atom(position=[0.5, 0.5, 0.5], symbols='In', name='In_d')
        structure.append_atom(position=[0.7896, 0.6234, 0.5], symbols='In', name='In_d')
        structure.append_atom(position=[0.75, 0.75, 0.75], symbols='As')
    elif request.param == 'str-Al':
        larray = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]])
        alat = 4.04
        structure = DataFactory('core.structure')(cell=larray * alat)
        structure.append_atom(position=np.array([0, 0, 0]) * alat, symbols='Al')
        structure.append_atom(position=np.array([0, 0.5, 0.5]) * alat, symbols='Al')
        structure.append_atom(position=np.array([0.5, 0, 0.5]) * alat, symbols='Al')
        structure.append_atom(position=np.array([0.5, 0.5, 0]) * alat, symbols='Al')
    elif request.param == 'str-InAs':
        structure_cls = DataFactory('core.structure')
        structure = structure_cls(cell=np.array([[0, 0.5, 0.5], [0.5, 0, 0.5], [0.5, 0.5, 0]]) * 6.058)
        structure.append_atom(position=(0, 0, 0), symbols='In', name='Hamburger')
        structure.append_atom(position=(0.25, 0.25, 0.25), symbols='As', name='Pizza')
    return structure


@pytest.fixture(params=['mesh', 'list'])
def vasp_kpoints(request, aiida_profile, data_path):
    """Fixture: (kpoints object, resulting KPOINTS)."""

    def _ref_kp_list():
        with open(data_path('kpoints', 'KPOINTS_list'), 'r', encoding='utf8') as reference_kpoints_fo:
            ref_kp_str = reference_kpoints_fo.read()
        return ref_kp_str

    def _ref_kp_mesh():
        with open(data_path('kpoints', 'KPOINTS_mesh'), 'r', encoding='utf8') as reference_kpoints_fo:
            ref_kp_list = reference_kpoints_fo.read()
        return ref_kp_list

    if request.param == 'mesh':
        kpoints = DataFactory('core.array.kpoints')()
        kpoints.set_kpoints_mesh([2, 2, 2])
        ref_kpoints = _ref_kp_mesh()
    elif request.param == 'list':
        kpoints = DataFactory('core.array.kpoints')()
        kpoints.set_kpoints([[0.0, 0.0, 0.0], [0.0, 0.0, 0.5]], weights=[1.0, 1.0])
        ref_kpoints = _ref_kp_list()
    return kpoints, ref_kpoints


@pytest.fixture(scope='session')
def potcar_family_name() -> str:
    """Return the POTCAR family name."""
    _potcar_family_name = 'test_family'
    return _potcar_family_name


@pytest.fixture(scope='session')
def potcar_mapping() -> dict:
    """Return the POTCAR mapping."""
    _potcar_mapping = {
        'In': 'In_sv',
        'In_d': 'In_d',
        'As': 'As',
        'Ga': 'Ga',
        'Si': 'Si',
        'P': 'P',
        'S': 'S',
        'Zn': 'Zn',
        'N': 'N',
        'H': 'H',
    }
    return _potcar_mapping


@pytest.fixture
def temp_pot_folder(tmp_path, data_path):
    """A temporary copy of the potcar test data folder, to avoid extracting tar objects inside the repo."""
    potcar_ga = pathlib.Path(data_path('potcar')) / 'Ga'
    assert not potcar_ga.exists()
    pot_archive = pathlib.Path(data_path('potcar'))
    target = tmp_path / 'potentials'
    # Ensure that the target path exists
    pathlib.Path(target).mkdir(exist_ok=True)
    copytree(pot_archive, target)
    return target


@pytest.fixture
def upload_potcar(aiida_profile, temp_pot_folder, potcar_family_name, data_path):
    """Upload a POTCAR family to DB."""
    if PotcarData.get_potcar_group(potcar_family_name) is not None:
        # Already uploaded, so we do not need to do it again.
        return

    def duplicate_potcar_data(potcar_node):
        """Create and store (and return) a duplicate of a given PotcarData node."""
        file_node = PotcarFileData()
        with temp_potcar(potcar_node.get_content()) as potcar_file:
            file_node.add_file(potcar_file)
            file_node.base.attributes.set('sha512', 'abcd')
            file_node.base.attributes.set('full_name', potcar_node.full_name)
            file_node.store()
        data_node, _ = PotcarData.get_or_create(file_node)
        return data_node

    potcar_ga = pathlib.Path(data_path('potcar')) / 'Ga'
    family_name = potcar_family_name
    family_desc = 'A POTCAR family used as a test fixture. Contains only unusable POTCAR files.'
    potcar_cls = PotcarData
    potcar_cls.upload_potcar_family(str(temp_pot_folder), family_name, family_desc, stop_if_existing=False)
    if len(potcar_cls.find(full_name='In_d')) == 1:
        family_group = potcar_cls.get_potcar_group(potcar_family_name)
        in_d = potcar_cls.find(full_name='In_d')[0]
        in_d_double = duplicate_potcar_data(in_d)
        family_group.add_nodes(in_d_double)
        assert in_d.uuid == potcar_cls.find(full_name='In_d')[0].uuid
    assert 'As' in potcar_cls.get_full_names(potcar_family_name, 'As')
    assert 'Ga' in potcar_cls.get_full_names(potcar_family_name, 'Ga')
    assert 'In_d' in potcar_cls.get_full_names(potcar_family_name, 'In')
    assert not potcar_ga.exists()


@pytest.fixture
def potentials(upload_potcar, potcar_family_name, potcar_mapping):
    """Fixture for two incomplete POTPAW potentials."""
    potcar_cls = PotcarData
    potentials = potcar_cls.get_potcars_dict(
        ['In', 'In_d', 'As'], family_name=potcar_family_name, mapping=potcar_mapping
    )

    return potentials


@pytest.fixture
def legacy_potcar_family(upload_potcar, potcar_family_name):
    """
    Fixture from creating an legacy potcar group

    Returns a tuple of group label and the LegacyPotcarGroup with the old type_string
    """

    class LegacyPotcarGroup(Group):
        """Old style group with the old type string"""

    # Override the _type_string class property which is supposed to be loaded from the entrypoint.
    LegacyPotcarGroup._type_string = OLD_POTCAR_FAMILY_TYPE
    new_group = PotcarGroup.collection.get(label=potcar_family_name)
    old_group = LegacyPotcarGroup(label=potcar_family_name + '_migrate_test')
    old_group.store()

    # Add the nodes from the new group to the old group
    old_group.add_nodes(list(new_group.nodes))
    return old_group.label, LegacyPotcarGroup


@pytest.fixture
def phonondb_run(tmp_path, data_path):
    phonondb = data_path('phonondb')
    copytree(phonondb, tmp_path)
    yield tmp_path


@pytest.fixture()
def run_vasp_process(
    aiida_profile,
    upload_potcar,
    vasp_params,
    potcar_family_name,
    potcar_mapping,
    vasp_kpoints,
    vasp_structure,
    mock_vasp,
):
    """Setup a standard VaspCalculation or VaspWorkChain with the mock executable that accepts input overrides."""

    def inner(inputs=None, settings=None, test_case=None, process_type='calcjob'):
        """
        Run a VaspCalculation or VaspWorkChain with specified input and settings overrides.

        Specific outputs can be selected using the test_case parameter.

        The type of process is set with the process_type parameter.
        """
        inpts = AttributeDict()
        inpts.structure = vasp_structure
        parameters = vasp_params.get_dict()
        options = {
            'withmpi': False,
            'queue_name': 'None',
            'resources': {'num_machines': 1, 'num_mpiprocs_per_machine': 1},
            'max_wallclock_seconds': 3600,
        }
        if test_case is not None:
            # Allow to fetch special tests cases using the mock-vasp executable
            parameters['system'] = f'test-case:{test_case}'
        if process_type == 'calcjob':
            process = CalculationFactory('vasp.vasp')
            inpts.potential = PotcarData.get_potcars_from_structure(
                structure=inpts.structure,
                family_name=potcar_family_name,
                mapping=potcar_mapping,
            )
            inpts.parameters = orm.Dict(dict=parameters)
            inpts.metadata = {}
            inpts.metadata['options'] = options
        elif process_type == 'workchain':
            process = WorkflowFactory('vasp.vasp')
            inpts.potential_family = orm.Str(potcar_family_name)
            inpts.potential_mapping = orm.Dict(dict=potcar_mapping)
            inpts.parameters = orm.Dict(dict={'incar': parameters})
            inpts.options = orm.Dict(dict=options)
            inpts.max_iterations = orm.Int(1)
            inpts.clean_workdir = orm.Bool(False)
            inpts.verbose = orm.Bool(True)
        else:
            raise ValueError(
                f"The supplied process_type: {process_type} is not supported. Use either 'calcjob' or 'workchain.'"
            )

        mock_vasp.store()
        # create_authinfo(computer=mock_vasp.computer, store=True)
        inpts.code = load_code('mock-vasp-loose@localhost')
        kpoints, _ = vasp_kpoints
        inpts.kpoints = kpoints
        if inputs is not None:
            # Allow overrides of the input
            inpts.update(inputs)
        if settings is not None and isinstance(settings, dict):
            inpts.settings = Dict(dict=settings)
        results_and_node = run.get_node(process, **inpts)
        return results_and_node

    return inner


@pytest.fixture()
def mock_potcars(aiida_profile, temp_pot_folder):
    """Create family of potcars for mock-vasp"""
    path = os.environ.get('MOCK_VASP_POTCAR_PATH', None)
    if path:
        PotcarData.upload_potcar_family(path, 'PBE.54', 'Family for mock calculation', stop_if_existing=False)
        return 'PBE.54'
    else:
        PotcarData.upload_potcar_family(
            str(temp_pot_folder), 'PBE.54', 'Family for mock calculation', stop_if_existing=False
        )
    return None


@pytest.fixture()
def builder_updater(
    aiida_profile,
    mock_potcars,
    mock_vasp,
):
    """
    Return a Builder Updater object for mock-vasp
    """
    return VaspBuilderUpdater(code='mock-vasp@localhost')


def print_and_export_failed_mock():
    """
    Print details about any failed mock
    """

    query = QueryBuilder()
    query.append(
        CalculationNode,
        filters={
            'or': [
                {'attributes.process_state': 'excepted'},
                {'attributes.exit_status': {'!==': 0}},
            ],
        },
        project=['*'],
    )
    query.append(Code, with_outgoing=CalculationNode, filters={'extras.is_mock_code': True})
    if query.count() == 0:
        return

    # Print information
    print('######## Information for FAILED mock code calculations ########')
    entities = []
    for (calcjob,) in query.all():
        entities.append(calcjob)
        root = get_call_root(calcjob)
        if root != calcjob:
            print('######## Information for the call root ########')
            print(format_call_graph(root))
            print(get_node_info(root))
            print(get_workchain_report(root, 'REPORT'))

        print('######## Information for the calcjob ########')
        print(get_node_info(calcjob))
        print(get_calcjob_report(calcjob))
        names = calcjob.outputs.retrieved.base.repository.list_object_names()
        stdout = None
        for option in ['stdout', 'vasp_output']:
            if option in names:
                stdout = option
                break
        if stdout is not None:
            print('######## STDOUT from the calcjob ########')
            print(calcjob.outputs.retrieved.base.repository.get_object_content(stdout))
        else:
            print('ERROR: No STDOUT found for the calculation')

    # Export the failed calculations
    output_file = cwd / 'test_mock_error.aiida'
    create_archive(entities, filename=str(output_file), include_logs=True, overwrite=True)


def get_call_root(node):
    """Obtain the root of the caller"""
    caller = node
    while True:
        next_caller = caller.caller

        if next_caller is None:
            break
        caller = next_caller
    return caller


def _create_mock_vasp_code(aiida_profile, localhost, exec_name):
    """
    Points to a mock-up of a VASP executable.

    If environmental variable REAL_VASP_PATH is set, switch the code
    to point to the REAL VASP executable. This is used to generate the
    actual outputs for mock tests later
    """
    query_builder = QueryBuilder()
    query_builder.append(Code, tag='code')
    query_builder.add_filter('code', {'label': {'==': exec_name}})
    query_results = query_builder.all()
    if query_results:
        code = query_results[0][0]
    else:
        os_env = os.environ.copy()
        if not localhost.pk:
            localhost.store()
        # returns unicode
        mock_vasp_path = sp.check_output(['which', exec_name], env=os_env, universal_newlines=True).strip()

        # Allow overriding mock using REAL code, this is used for running the actual
        # calculation once and deposit the results in the registry
        if os.environ.get('REAL_VASP_PATH'):
            mock_vasp_path = os.environ['REAL_VASP_PATH']

        code = InstalledCode(localhost, mock_vasp_path)
        code.label = exec_name
        code.description = 'Mock VASP for tests'
        code.default_calc_job_plugin = 'vasp.vasp'
        code.store()
        code.base.extras.set('is_mock_code', True)

    return code



"""Test extended dicts."""
# pylint: disable=unused-import,redefined-outer-name,unused-argument,unused-wildcard-import,wildcard-import,no-member

import pytest
from aiida.common.extendeddicts import AttributeDict

from aiida_vasp.utils.extended_dicts import delete_keys_from_dict, update_nested_dict


@pytest.fixture
def init_dict1():
    """Fixture for dictionary one."""
    dct = AttributeDict()
    dct.dct1 = AttributeDict()
    dct.dct1.test = 1.0
    dct.dct1.test2 = 'string1'
    dct.dct1.test3 = [1.0]
    dct.dct1.test4 = {'key1': 'value1'}

    return dct


@pytest.fixture
def init_dict2():
    """Fixture for dictionary two."""
    dct = AttributeDict()
    dct.dct1 = AttributeDict()
    dct.dct1.test = 2.0
    dct.dct1.test2 = 'string2'
    dct.dct1.test3 = [2.0]
    dct.dct1.test4 = {'key2': 'value2'}

    return dct


def test_nested_dict_update(init_dict1, init_dict2):
    """Test a nested dict update."""
    dict1 = AttributeDict(init_dict1)
    dict2 = AttributeDict(init_dict2)
    update_nested_dict(dict1, dict2)
    test = AttributeDict(
        {
            'dct1': AttributeDict(
                {
                    'test': 2.0,
                    'test2': 'string2',
                    'test3': [2.0],
                    'test4': {'key1': 'value1', 'key2': 'value2'},
                }
            )
        }
    )
    assert test == dict1
    dict1 = AttributeDict(init_dict1)
    del dict2.dct1.test
    update_nested_dict(dict1, dict2)
    test.dct1.test = 1.0
    assert test == dict1


def test_nested_dict_delete(init_dict1):
    """Test nested dict delete."""
    dict1 = AttributeDict(init_dict1)
    delete_keys_from_dict(dict1, 'dct1.test')
    test = AttributeDict(
        {
            'dct1': AttributeDict(
                {
                    'test2': 'string1',
                    'test3': [1.0],
                    'test4': AttributeDict({'key1': 'value1'}),
                }
            )
        }
    )
    assert test == dict1
    delete_keys_from_dict(dict1, 'dct1.test5')
    assert test == dict1
    delete_keys_from_dict(dict1, 'dct2')
    assert test == dict1



from .base import BaseFileParser
from .chgcar import ChgcarParser
from .doscar import DoscarParser
from .eigenval import EigenvalParser
from .kpoints import KpointsParser
from .outcar import OutcarParser, VtstNebOutcarParser
from .poscar import PoscarParser
from .stream import StreamParser
from .vasprun import VasprunParser

__all__ = (
    'BaseFileParser',
    'ChgcarParser',
    'DoscarParser',
    'EigenvalParser',
    'KpointsParser',
    'OutcarParser',
    'PoscarParser',
    'StreamParser',
    'VasprunParser',
    'VtstNebOutcarParser',
)



import pytest
from aiida import orm

from aiida_vasp.utils.opthold import BandOptions, ConvOptions, OptionContainer, RelaxOptions


def test_option_container():
    """Test the base option container"""

    class TestHolder(OptionContainer):
        x: int = 1
        y: str = 'test'
        z: float

    holder = TestHolder(z=1.0)
    d = holder.aiida_dict()
    holder2 = TestHolder(**d.get_dict())
    assert holder == holder2
    assert holder.model_dump() == holder2.model_dump()

    assert holder.aiida_validate({'x': 2, 'y': 'test2', 'z': 2.0}) is None
    assert holder.aiida_validate({'x': 2, 'y': 2.0, 'z': '2'}) is not None

    node = holder.aiida_serialize({'x': 2, 'z': 3})
    assert node['x'] == 2
    assert node['y'] == 'test'
    assert node['z'] == 3


@pytest.mark.parametrize('cls', [RelaxOptions, BandOptions, ConvOptions])
def test_workchain_options_roundtrip(cls):
    """Test the relax option container"""
    opt = cls()
    out = opt.aiida_dict()
    assert cls(**out.get_dict()) == opt


@pytest.mark.parametrize('cls', [RelaxOptions, BandOptions, ConvOptions])
def test_workchain_options_validate(cls):
    """Test the relax option container"""
    opt = cls()
    out = opt.model_dump()
    assert cls.aiida_validate(out) is None


@pytest.mark.parametrize('cls', [RelaxOptions, BandOptions, ConvOptions])
def test_workchain_options_serialize(cls):
    """Test the relax option container"""
    opt = cls()
    out = opt.model_dump()
    assert isinstance(cls.aiida_serialize(out), orm.Dict)


@pytest.mark.parametrize('cls', [RelaxOptions, BandOptions, ConvOptions])
def test_workchain_options_description(cls):
    """Test the relax option container"""
    opt = cls()
    desc = opt.aiida_description()
    assert desc



"""
Utilities for Wannier90.

Utility to convert raw input data to .win format.
"""

from typing import Any, Sequence


class DictToWin(object):  # pylint: disable=useless-object-inheritance
    """Format parameters given in a dictionary into Wannier90 .win format."""

    @classmethod
    def _bool(cls, val: Any) -> bool:
        return 'T' if val else 'F'

    @classmethod
    def _seq(cls, val: Sequence) -> str:
        """String format a sequence."""
        res = []
        for i in val:
            if not isinstance(i, (list, tuple)):
                line = cls._value(i)
            else:
                line = ' '.join(map(cls._value, i))
            res.append(' ' + line)
        return res

    @classmethod
    def _block(cls, name: str, val: Any) -> list[str]:
        """
        Create a win parameter block with name being the block name, val being the content in python representation.
        """
        res = ['begin ' + name]
        res += cls._value(val)
        res += ['end ' + name]
        return res

    @classmethod
    def _assign(cls, key: str, val: Any) -> str:
        return f'{key} = {val}'

    @classmethod
    def _value(cls, val: Any) -> str:
        """String format a value of any compatible scalar type."""
        if isinstance(val, str):  # pylint: disable=undefined-variable
            return val
        if isinstance(val, bool):
            return cls._bool(val)
        if isinstance(val, (list, tuple)):
            return cls._seq(val)
        return str(val)

    @classmethod
    def _item(cls, key: str, val: Any) -> list[str]:
        if isinstance(val, (list, tuple)):
            return cls._block(key, val)
        return [cls._assign(key, cls._value(val))]

    @classmethod
    def parse(cls, in_dict: dict) -> str:
        """Parse a dictionary into win formatted string."""
        res = []
        for key, value in in_dict.items():
            res += cls._item(key, value)
        return '\n'.join(res)



"""
Module for using pymatgen.io.vasp.sets based input sets.
"""

from typing import Dict, Optional, Union

import numpy as np
from aiida import orm

from .base import InputSet

try:
    import pymatgen.io.vasp.sets as pmg_sets
    from pymatgen.io.vasp.inputs import KpointsSupportedModes
except ImportError:
    pmg_sets = None


class PymatgenInputSet(InputSet):
    """
    Input set using pymatgen.io.vasp.sets.

    Provides basic compatibility with pymatgen sets for generating VASP input
    parameters, k-point meshes, and pseudopotential mappings.
    """

    # An none-exhaust list of supported pymatgen input sets
    KNOWN_SETS = (
        'MPRelaxSet',
        'MITRelaxSet',
        'MPScanRelaxSet',
        'MP24RelaxSet',
        'MPMetalRelaxSet',
        'MPHSERelaxSet',
        'MVLGWSet',
        'MPAbsorptionSet',
        'MatPESStaticSet',
        'MPScanStaticSet',
        'MP24StaticSet',
        'MPHSEBSSet',
        'MPNonSCFSet',
        'MPSOCSet',
        'MPNMRSet',
        'MPStaticSet',
    )

    def __init__(
        self,
        set_name: str,
        overrides: Optional[Dict] = None,
        verbose: Optional[bool] = None,
        pmg_kwargs: Optional[Dict] = None,
    ) -> None:
        """
        Instantiate a PymatgenInputSet.

        :param set_name: Name of the pymatgen input set to use
        :type set_name: str
        :param overrides: Dictionary of parameter overrides for the input set
        :type overrides: dict or None
        :param verbose: If True, print additional information during processing
        :type verbose: bool or None
        :param pmg_kwargs: Additional keyword arguments to pass to the pymatgen input set
        :type pmg_kwargs: dict or None

        :raises AssertionError: If set_name is not in KNOWN_SETS
        """
        assert set_name in self.KNOWN_SETS, f'Unsupported set name: {set_name}'
        super().__init__(set_name, overrides=overrides, verbose=verbose)
        self._pmg_kwargs = pmg_kwargs or {}

    def _load_data(self) -> None:
        """
        Load the pymatgen input set class.

        Dynamically imports and stores the pymatgen input set class based on set_name.

        :raises ImportError: If pymatgen is not installed or cannot be imported
        """
        if pmg_sets is None:
            raise ImportError('pymatgen is not installed. Please install it to use PymatgenInputSet.')
        self._pmg_class = getattr(pmg_sets, self.set_name)

    def get_input_dict(self, structure: orm.StructureData, raw_python: bool = True) -> Union[Dict, orm.Dict]:
        """
        Compute the input parameters for a VASP calculation using pymatgen.io.vasp.sets.

        Generates INCAR parameters by instantiating the pymatgen input set with the
        given structure and applying any specified overrides. Removes certain
        parameters that conflict with aiida-vasp's input validation.

        :param structure: Crystal structure for the calculation
        :type structure: orm.StructureData
        :param raw_python: If True, return a Python dict; if False, return orm.Dict
        :type raw_python: bool

        :returns: Dictionary of INCAR parameters
        :rtype: dict or orm.Dict
        """
        ps = structure.get_pymatgen()
        pmgset = self._pmg_class(ps, **self._pmg_kwargs)
        incar_dict = {key.lower(): value for key, value in pmgset.incar.items()}
        # Apply the overrides
        for key, value in self.overrides.items():
            if value is None:
                if key in incar_dict:
                    incar_dict.pop(key)
            else:
                incar_dict[key] = value

        # pop icharg which conflicts with aiida-vasp's input checks
        incar_dict.pop('icharg', None)
        incar_dict.pop('istart', None)
        incar_dict.pop('kspacing', None)

        if raw_python:
            return incar_dict
        return orm.Dict(dict=incar_dict)

    def get_pp_mapping(self, structure: orm.StructureData) -> Dict[str, str]:
        """
        Get the pseudopotential mapping used by the input set.

        Returns a dictionary mapping element symbols to their corresponding
        pseudopotential symbols as defined by the pymatgen input set.

        :param structure: Crystal structure for the calculation
        :type structure: orm.StructureData

        :returns: Dictionary mapping element names to pseudopotential symbols
        :rtype: dict
        """
        ps = structure.get_pymatgen()
        pmgset = self._pmg_class(ps, **self._pmg_kwargs)
        return {p.element: p.symbol for p in pmgset.potcar}

    def get_potcar_family(self) -> str:
        """
        Get the POTCAR family used by the input set.

        Retrieves the pseudopotential functional family from the pymatgen
        input set configuration. Converts underscore notation to dot notation
        (e.g., PBE_54 becomes PBE.54) for aiida-vasp compatibility.

        :returns: Name of the POTCAR family
        :rtype: str
        """
        return self._pmg_class.CONFIG['POTCAR_FUNCTIONAL'].replace('_', '.')

    def get_kpoints(self, structure: orm.StructureData) -> Optional[orm.KpointsData]:
        """
        Return a KpointsData object for the given structure.

        Converts the k-point specification from the pymatgen input set to an
        aiida-vasp compatible KpointsData object. Supports Gamma-centered,
        Monkhorst-Pack, and automatic k-point generation modes.

        :param structure: Crystal structure for k-point generation
        :type structure: orm.StructureData

        :returns: K-points data object, or None if no k-points are specified
        :rtype: orm.KpointsData or None
        """
        ps = structure.get_pymatgen()
        pmgset = self._pmg_class(ps, **self._pmg_kwargs)
        if pmgset.kpoints is None:
            return None
        # Currently only supports Gamma and Monkhorst-Pack
        kpoints_data = pmg_kpoints2kpointsdata(pmgset.kpoints, structure)
        return kpoints_data

    def get_kpoints_spacing(self, structure: orm.StructureData) -> Optional[float]:
        """
        Get the k-point spacing used by the input set.

        Extracts the KSPACING parameter from the pymatgen input set and converts
        it to the format expected by aiida-vasp (dividing by 2π).

        :param structure: Crystal structure for the calculation
        :type structure: orm.StructureData

        :returns: K-point spacing value or None if not specified
        :rtype: float or None
        """
        ps = structure.get_pymatgen()
        pmgset = self._pmg_class(ps, **self._pmg_kwargs)
        incar_dict = {key.lower(): value for key, value in pmgset.incar.items()}
        kspacing = incar_dict.pop('kspacing', None)
        if kspacing is not None:
            return kspacing / np.pi / 2
        return None


def pmg_kpoints2kpointsdata(pmg_kpoints, structure: orm.StructureData) -> orm.KpointsData:
    """
    Convert a pymatgen Kpoints object to an AiiDA KpointsData object.

    Handles conversion between different k-point generation modes:
    - Gamma-centered grids
    - Monkhorst-Pack grids (with appropriate shift corrections)
    - Automatic k-point generation based on spacing

    :param pmg_kpoints: Pymatgen Kpoints object to convert
    :type pmg_kpoints: pymatgen.io.vasp.inputs.Kpoints
    :param structure: AiiDA structure data for setting the unit cell
    :type structure: orm.StructureData

    :returns: Converted k-points data object
    :rtype: orm.KpointsData

    :raises ValueError: If the k-point style is not supported
    """

    # Currently only supports Gamma and Monkhorst-Pack
    style = pmg_kpoints.style
    mesh = pmg_kpoints.kpts[0]
    kpoints = orm.KpointsData()

    if style == KpointsSupportedModes.Gamma:
        # aiida-vasp defaults to use Gamma-centering mode when constructing the KPOINTS file
        gamma_shifts = (0, 0, 0)
    elif style == KpointsSupportedModes.Monkhorst:
        # If a MP grid is supplied, we just add a -0.5 shift to the gamma-centering grid to make it
        # equivalent to a MP centred grid
        # See https://www.vasp.at/wiki/index.php/KPOINTS
        gamma_shifts = []
        for i in mesh:
            if i % 2 == 1:
                # Odd division - do nothing
                gamma_shifts.append(0)
            else:
                # Even division - add shifts
                gamma_shifts.append(-0.5)
    elif style == KpointsSupportedModes.Automatic:
        # The automatic mode is used with a length R_k
        # We convert it to a mesh based on the structure's lattice
        # See: https://www.vasp.at/wiki/index.php/KPOINTS#Automatic_k-point_mesh
        kspacing = 2 * np.pi / pmg_kpoints.kpts[0][0]
        kpoints.set_cell_from_structure(structure)
        kpoints.set_kpoints_mesh_from_density(kspacing)
        return kpoints
    else:
        raise ValueError(f'Unsupported kpoint style: {style}')
    # Using explicit meshes
    shifts = pmg_kpoints.kpts_shift
    # Construct AiiDA KpointsData object
    kpoints.set_kpoints_mesh(mesh, offset=[i + j for i, j in zip(shifts, gamma_shifts)])
    kpoints.set_cell_from_structure(structure)
    return kpoints



"""Unit tests for mock-vasp command."""

# pylint: disable=unused-import,unused-argument,redefined-outer-name
import subprocess


def test_mock_vasp():
    """A simple test to verify that mock-vasp can run."""
    subprocess.check_output(['mock-vasp', '--help'])



"""Test aiida_utils functionss."""

import warnings

import pytest

from aiida_vasp.utils.aiida_utils import convert_dict_case, get_current_user

try:
    from aiida_vasp.utils.pmg import PymatgenAdapator, get_incar, get_kpoints, get_outcar, get_vasprun
except ImportError:
    PMG_INSTALLED = False
else:
    PMG_INSTALLED = True


def test_get_current_user(aiida_profile_clean):
    """Assert that get_current_user returns a user."""
    user = get_current_user()
    assert user.pk
    assert user.first_name == ''
    assert user.last_name == ''
    assert user.email


@pytest.mark.skipif(not PMG_INSTALLED, reason='pymatgen not installed')
@pytest.mark.parametrize(['vasp_structure', 'vasp_kpoints'], [('str', 'mesh')], indirect=True)
def test_pmg_adaptor(aiida_profile_clean, tmp_path, run_vasp_process):
    """
    Test export vasp calculation
    """

    _, node = run_vasp_process()

    adapt = PymatgenAdapator(node)
    assert adapt.vasprun

    assert 'vasprun' in adapt.pmg_objects

    with PymatgenAdapator(node) as adapt:
        adapt.vasprun

    assert 'vasprun_dict' in adapt.cache
    assert 'pmg_cache' in node.base.extras.all

    assert get_incar(node)
    assert get_kpoints(node)
    assert get_vasprun(node)
    assert get_outcar(node)


def test_convert_dict_case_lowercase():
    input_dict = {'KeyOne': 1, 'KEYTWO': 2}
    expected_output = {'keyone': 1, 'keytwo': 2}
    assert convert_dict_case(input_dict) == expected_output


def test_convert_dict_case_uppercase():
    input_dict = {'keyone': 1, 'keytwo': 2}
    expected_output = {'KEYONE': 1, 'KEYTWO': 2}
    assert convert_dict_case(input_dict, lower=False) == expected_output


def test_convert_dict_case_nested():
    input_dict = {'KeyOne': {'SubKeyOne': 1, 'SUBKEYTWO': 2}, 'KEYTWO': 2}
    expected_output = {'keyone': {'subkeyone': 1, 'subkeytwo': 2}, 'keytwo': 2}
    assert convert_dict_case(input_dict) == expected_output


def test_convert_dict_case_warning(caplog):
    input_dict = {'KeyOne': 1, 'KEYTWO': 2}
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter('always')
        convert_dict_case(input_dict, warn=True)
        assert len(w) == 2
        assert "Key 'KeyOne' converted to 'keyone' - please use lower case keys" in str(w[0].message)
        assert "Key 'KEYTWO' converted to 'keytwo' - please use lower case keys" in str(w[1].message)


def test_convert_dict_case_raise():
    input_dict = {'KeyOne': 1, 'KEYTWO': 2}
    with pytest.raises(ValueError, match="Key 'KeyOne' converted to 'keyone' - please use lower case keys"):
        convert_dict_case(input_dict, raise_convert=True)


def test_convert_dict_case_no_conversion():
    input_dict = {'keyone': 1, 'keytwo': 2}
    assert convert_dict_case(input_dict) == input_dict



"""
The eos workchain

----------------

The workchain will accept a dictionary of structures and extract the
total energies for each structure.
The data is saved and the energy minimum is calculated and stored.
"""
# pylint: disable=attribute-defined-outside-init
import random

import numpy as np
from aiida.common.extendeddicts import AttributeDict
from aiida.engine import WorkChain, append_, calcfunction, while_
from aiida.plugins import DataFactory, WorkflowFactory
from aiida_vasp.utils.workchains import compose_exit_code, prepare_process_inputs
from scipy.interpolate import interp1d
from scipy.optimize import minimize


class EosWorkChain(WorkChain):
    """
    The eos workchain

    ----------------

    The workchain will accept a dictionary of structures and extract the
    total energies for each structure.
    The data is saved and the energy minimum is calculated and stored.
    """

    _verbose = False
    _next_workchain_string = 'vasp.vasp'
    _next_workchain = WorkflowFactory(_next_workchain_string)

    @classmethod
    def define(cls, spec):
        super(EosWorkChain, cls).define(spec)
        spec.expose_inputs(cls._next_workchain, exclude=['structure'])
        spec.input_namespace('structures', valid_type=DataFactory('structure'), dynamic=True, help='a dictionary of structures to use')
        spec.exit_code(0, 'NO_ERROR', message='the sun is shining')
        spec.exit_code(420, 'ERROR_NO_CALLED_WORKCHAIN', message='no called workchain detected')
        spec.exit_code(500, 'ERROR_UNKNOWN', message='unknown error detected in the eos workchain')

        spec.outline(
            cls.initialize,
            while_(cls.run_next_workchains)(
                cls.init_next_workchain,
                cls.run_next_workchain,
                cls.verify_next_workchain,
                cls.extract_volume_and_energy
            ),
            cls.finalize
        )  # yapf: disable

        spec.output('eos', valid_type=DataFactory('array'), help='a list containing the cell volumes and total energies')
        spec.output('eos_minimum', valid_type=DataFactory('dict'), help='a dictionary containing the cell volume at energy minimum')

    def initialize(self):
        """Initialize the eos workchain."""
        self._init_context()
        self._init_inputs()

    def _init_context(self):
        """Initialize context variables that are used during the."""
        # Set the exit code to error in case we forget to set it to NO_ERROR
        self.ctx.exit_code = self.exit_codes.ERROR_UNKNOWN  # pylint: disable=no-member

        # Copy structures to context, since we will empty it as we go.
        # Since structures is an input to a workchain we cannot modify it and need to copy.
        self.ctx.structures = dict(self.inputs.structures)

        # Continue to submit workchains until this is True
        self.ctx.is_finished = False

        # Define an interation index
        self.ctx.iteration = 0

        # Define the context inputs
        self.ctx.inputs = AttributeDict()

        # Define the total energies list
        self.ctx.total_energies = []

    def _init_inputs(self):
        """Initialize inputs."""

        # Set verbose flag (more or less details on the workchain report)
        try:
            self._verbose = self.inputs.verbose.value
        except AttributeError:
            pass

    def run_next_workchains(self):
        """
        Return whether a new workchain should be run.

        This is the case as long as the last workchain has not finished successfully.
        """
        return not self.ctx.is_finished

    def init_next_workchain(self):
        """Initialize the next workchain."""

        # Elevate iteration index
        self.ctx.iteration += 1

        # Check that the context inputs exists
        try:
            self.ctx.inputs
        except AttributeError:
            raise ValueError('No input dictionary was defined in self.ctx.inputs')

        # Take the exposed inputs and add them to the context input. This would
        # be the inputs you supply to this workchain
        self.ctx.inputs.update(self.exposed_inputs(self._next_workchain))

        # We did not expose the structure as we would like to set this
        # from the supplied structures input. Just choose any item in the
        # structures dictionary and asign that to the next run.
        item = random.choice(list(self.ctx.structures.keys()))
        self.ctx.inputs.structure = self.ctx.structures.pop(item)

        # Make sure we do not have any floating dict (convert to Dict etc.)
        self.ctx.inputs = prepare_process_inputs(self.ctx.inputs, namespaces=['dynamics'])

    def run_next_workchain(self):
        """
        Run the next workchain

        It is either submitted to the daemon or run, depending on how you
        run this workchain. The execution method is inherited.
        """
        inputs = self.ctx.inputs
        running = self.submit(self._next_workchain, **inputs)
        self.report(f'launching {self._next_workchain.__name__}<{running.pk}> iteration #{self.ctx.iteration}')
        self.to_context(workchains=append_(running))

    def verify_next_workchain(self):
        """Correct for unexpected behavior."""

        try:
            workchain = self.ctx.workchains[-1]
        except IndexError:
            self.report(f'There is no {self._next_workchain.__name__} in the called workchain list.')
            return self.exit_codes.ERROR_NO_CALLED_WORKCHAIN  # pylint: disable=no-member

        # Inherit exit status from last workchain (supposed to be
        # successfull)
        next_workchain_exit_status = workchain.exit_status
        next_workchain_exit_message = workchain.exit_message
        if not next_workchain_exit_status:
            self.ctx.exit_code = self.exit_codes.NO_ERROR  # pylint: disable=no-member
        else:
            self.ctx.exit_code = compose_exit_code(next_workchain_exit_status, next_workchain_exit_message)
            self.report(
                'The called {}<{}> returned a non-zero exit status. ' 'The exit status {} is inherited'.format(
                    workchain.__class__.__name__, workchain.pk, self.ctx.exit_code
                )
            )

        # Stop further execution of workchains if there are no more structure
        # entries in the structures dictionary
        if not self.ctx.structures:
            self.ctx.is_finished = True

        return self.ctx.exit_code

    def extract_volume_and_energy(self):
        """Extract the cell volume and total energy for this structure."""

        workchain = self.ctx.workchains[-1]

        # Fetch the total energy
        misc = workchain.outputs.misc.get_dict()
        total_energy = misc['total_energies']['energy_extrapolated']

        # Fetch the volume
        volume = self.ctx.inputs.structure.get_cell_volume()

        # Store both in a list
        self.ctx.total_energies.append([volume, total_energy])

    def finalize(self):
        """
        Finalize the workchain.

        Take the total energies container and set is as an output of this workchain.
        """
        # Due to data provenance we cannot return AiiDA data containers that have
        # not been passed through a calcfunction, workfunction or a workchain. Create this now.
        total_energies = store_total_energies(DataFactory('list')(list=self.ctx.total_energies))

        # Let us also try to find a better minimum, just as an example using the power of Python
        energy = locate_minimum(total_energies)

        # And then store the output on the workchain
        self.out('eos', total_energies)
        self.out('eos_minimum', energy)


@calcfunction
def store_total_energies(total_energies):
    """Stores the total energies in ArrayData to keep data provenance."""
    total_energies_list = total_energies.get_list()
    # Let us also sort by volume as we picked the entries in the structures by random
    # above
    total_energies_array = np.array(total_energies_list)
    total_energies_array_sorted = total_energies_array[total_energies_array[:, 0].argsort()]
    array_data = DataFactory('array')()
    array_data.set_array('eos', total_energies_array_sorted)

    return array_data


@calcfunction
def locate_minimum(total_energies):
    """Locate the volume with the lowest energy using interpolation."""
    total_energies_array = total_energies.get_array('eos')
    volumes = total_energies_array[:, 0]
    energies = total_energies_array[:, 1]

    # Establish some initial guess (take lowest energy point from the original dataset)
    min_energy_guess_index = np.argmin(energies)

    # Create the function that can be used to extract interpolated values
    # Using cubic interpolation here, which is not necessarly physically correct,
    # only serves as an example. Please have a look at the theory behind more realiastic
    # models that can be used to fit such data.
    new_energies = interp1d(volumes, energies, kind='cubic')

    # Use minimize from scipy to locate the minimal point that can be ejected by the
    # interpolation routines
    minimized_point = minimize(new_energies, volumes[min_energy_guess_index], tol=1e-3)

    # Create a dictionary to house the results and return
    dict_data = DataFactory('dict')(dict={'volume': minimized_point.x[0], 'energy': minimized_point.fun})

    return dict_data



"""
This file contains the VaspWorkChain class definition which uses the BaseRestartWorkChain.

Below is a copy of the error handler logic from aiida-core:

If the process is excepted or killed, the work chain will abort. Otherwise any attached handlers will be called
in order of their specified priority. If the process was failed and no handler returns a report indicating that
the error was handled, it is considered an unhandled process failure and the process is relaunched. If this
happens twice in a row, the work chain is aborted. In the case that at least one handler returned a report the
following matrix determines the logic that is followed::

    Process  Handler    Handler     Action
    result   report?    exit code

    -----------------------------------------

    Success      yes        == 0     Restart
    Success      yes        != 0     Abort
    Failed       yes        == 0     Restart
    Failed       yes        != 0     Abort

If no handler returned a report and the process finished successfully, the work chain's work is considered done
and it will move on to the next step that directly follows the `while` conditional, if there is one defined in
the outline.

This means that for a handler:

- No error found - just return None
- No action taken

  - the error is not recoverable - return with a non-zero error code with do break
  - the error is not recoverable, but other handler may/maynot save it - return with a non-zero code without do
    break
  - the error is not recoverable, and the workchain should be aborted immediately - non-zero code + do break

- Action taken

  - the error is fixed in full - return with a zero error code with `do_break=True`
  - the error is not fixed in full - return with a report with `do_break=False` but has a `exit_code`.
    this mean other handlers (with lower priority) must handle it and return a zero error_code.

"""

from __future__ import annotations

import math
from types import TracebackType
from typing import Any, List, Optional, Tuple

import numpy as np
from aiida import orm
from aiida.common.exceptions import InputValidationError, NotExistent

# from aiida.engine.job_processes import override
from aiida.common.extendeddicts import AttributeDict
from aiida.common.lang import override
from aiida.engine import ExitCode, ProcessSpec, if_, while_
from aiida.engine.processes.workchains.restart import (
    BaseRestartWorkChain,
    ProcessHandlerReport,
    WorkChain,
    process_handler,
)
from aiida.orm import CalcJobNode, Code, Dict, KpointsData
from aiida.orm.nodes.data.base import to_aiida_type
from aiida.plugins import CalculationFactory

from aiida_vasp.assistant.parameters import (
    ParametersMassage,
    inherit_and_merge_parameters,
)
from aiida_vasp.calcs.vasp import VaspCalculation
from aiida_vasp.data.chargedensity import ChargedensityData
from aiida_vasp.data.potcar import PotcarData
from aiida_vasp.data.wavefun import WavefunData
from aiida_vasp.utils.workchains import compose_exit_code, prepare_process_inputs, site_magnetization_to_magmom

from .common import parameters_validator
from .common.dryrun import get_jobscheme
from .inputset.vaspsets import get_ldau_keys
from .mixins import WithBuilderUpdater

# pylint: disable=no-member


class VaspWorkChain(BaseRestartWorkChain, WithBuilderUpdater):
    """
    The VASP workchain.

    Error handling enriched wrapper around VaspCalculation.

    Deliberately conserves most of the interface (required inputs) of the VaspCalculation class, but
    makes it possible for a user to interact with a workchain and not a calculation.

    This is intended to be used instead of directly submitting a VaspCalculation,
    so that future features like
    automatic restarting, error checking etc. can be propagated to higher level workchains
    automatically by implementing them here.

    Handlers are implemented to try fix common problems and improves the robustness.
    Individual handlers can be enabled/disabled by setting the ``handler_overrides`` input port.
    Additional settings may be passed under the "settings" input, which is also forwarded to the
    calculations. The available options are:

    - ``USE_WAVECAR_FOR_RESTART`` wether calculation restarts should use the WAVECAR. The default is ``True``.

    Usage::

        from aiida.common.extendeddicts import AttributeDict
        from aiida.work import submit
        basevasp = WorkflowFactory('vasp.vasp')
        inputs = basevasp.get_builder()
        inputs = AttributeDict()
        ## ... set inputs
        submit(basevasp, **inputs)

    To see a working example, including generation of input nodes from scratch, please
    refer to ``examples/run_vasp_lean.py``.
    """

    _verbose = False

    _process_class = CalculationFactory('vasp.vasp')
    _algo_switching = {
        'normal': ['fast', 'veryfast', 'damped'],
        'fast': ['normal', 'veryfast', 'damped'],
        'veryfast': ['normal', 'fast', 'damped'],
        'damped': ['normal', 'fast', 'veryfast'],
    }

    @classmethod
    def define(cls, spec: ProcessSpec) -> None:  # pylint: disable=too-many-statements
        super(VaspWorkChain, cls).define(spec)
        spec.input('code', valid_type=Code)
        spec.input(
            'structure',
            valid_type=(orm.StructureData, orm.CifData),
            required=True,
        )
        spec.input('kpoints', valid_type=orm.KpointsData, required=False)
        spec.input(
            'potential_family',
            valid_type=orm.Str,
            required=True,
            serializer=to_aiida_type,
            validator=potential_family_validator,
        )
        spec.input(
            'potential_mapping',
            valid_type=orm.Dict,
            required=True,
            serializer=to_aiida_type,
        )
        spec.input(
            'parameters',
            valid_type=orm.Dict,
            required=True,
            validator=parameters_validator,
            serializer=to_aiida_type,
        )
        spec.input(
            'options',
            valid_type=orm.Dict,
            required=True,
            serializer=to_aiida_type,
        )
        spec.input(
            'settings',
            valid_type=orm.Dict,
            required=False,
            serializer=to_aiida_type,
        )
        spec.input('wavecar', valid_type=WavefunData, required=False)
        spec.input('chgcar', valid_type=ChargedensityData, required=False)
        spec.input(
            'site_magnetization',
            valid_type=orm.Dict,
            required=False,
            help='Site magnetization to be used as MAGMOM',
        )
        spec.input(
            'restart_folder',
            valid_type=orm.RemoteData,
            required=False,
            help="""
            The restart folder from a previous workchain run that is going to be used.
            """,
        )
        spec.input(
            'max_iterations',
            valid_type=orm.Int,
            required=False,
            default=lambda: orm.Int(5),
            serializer=to_aiida_type,
            help="""
            The maximum number of iterations to perform.
            """,
        )
        spec.input(
            'clean_workdir',
            valid_type=orm.Bool,
            required=False,
            serializer=to_aiida_type,
            default=lambda: orm.Bool(True),
            help="""
            If True, clean the work dir upon the completion of a successfull calculation.
            """,
        )
        spec.input(
            'keep_last_workdir',
            valid_type=orm.Bool,
            default=lambda: orm.Bool(False),
            serializer=to_aiida_type,
            help='If True, prevent the last workdir from being cleaned in case the files are needed for restarts.',
        )
        spec.input(
            'verbose',
            valid_type=orm.Bool,
            required=False,
            serializer=to_aiida_type,
            default=lambda: orm.Bool(False),
            help="""
            If True, enable more detailed output during workchain execution.
            """,
        )
        spec.input(
            'ldau_mapping',
            valid_type=orm.Dict,
            required=False,
            serializer=to_aiida_type,
            help="""Settings for assign LDA+U related settings according to the input structure.

    mapping: a dictionary in the format of  {"Mn": [d, 4]...} for U
    utype: the type of LDA+U, default to 2, which is the one with only one parameter
    jmapping: a dictionary in the format of  {"Mn": [d, 4]...} but for J
    felec: Wether we are dealing with f electrons, will increase lmaxmix if we are.""",
        )
        spec.input(
            'kpoints_spacing',
            valid_type=orm.Float,
            required=False,
            serializer=to_aiida_type,
            help='Spacing for the kpoints in units A^-1 * 2pi',
        )
        spec.input(
            'auto_parallel',
            valid_type=orm.Dict,
            serializer=to_aiida_type,
            required=False,
            help='Automatic parallelisation settings, keywords passed to `get_jobscheme` function.',
        )
        spec.input(
            'dynamics.positions_dof',
            valid_type=orm.List,
            serializer=to_aiida_type,
            required=False,
            help="""
            Site dependent flag for selective dynamics when performing relaxation
            """,
        )
        spec.input(
            'vdw_kernel',
            valid_type=orm.SinglefileData,
            required=False,
            help='The vdw_kerenl.bindat file to be used for vdw calculations.',
        )
        spec.outline(
            cls.setup,
            cls.init_inputs,
            if_(cls.run_auto_parallel)(cls.prepare_inputs, cls.perform_autoparallel),
            while_(cls.should_run_process)(
                cls.prepare_inputs,
                cls.run_process,
                cls.inspect_process,
            ),
            cls.results,
        )  # yapf: disable
        spec.output('parallel_settings', valid_type=orm.Dict, required=False)
        spec.expose_outputs(cls._process_class)

        # Copied from the old plugin restart workchain
        spec.exit_code(
            0,
            'NO_ERROR',
            message='the sun is shining',
        )
        spec.exit_code(
            300,
            'ERROR_MISSING_REQUIRED_OUTPUT',
            message='the calculation is missing at least one required output in the restart workchain',
        )
        spec.exit_code(
            400,
            'ERROR_ITERATION_RETURNED_NO_CALCULATION',
            message='the run_calculation step did not successfully add a calculation node to the context',
        )
        spec.exit_code(
            401,
            'ERROR_MAXIMUM_ITERATIONS_EXCEEDED',
            message='the maximum number of iterations was exceeded',
        )
        spec.exit_code(
            402,
            'ERROR_UNEXPECTED_CALCULATION_STATE',
            message='the calculation finished with an unexpected calculation state',
        )
        spec.exit_code(
            403,
            'ERROR_UNEXPECTED_CALCULATION_FAILURE',
            message='the calculation experienced and unexpected failure',
        )
        spec.exit_code(
            404,
            'ERROR_SECOND_CONSECUTIVE_SUBMISSION_FAILURE',
            message='the calculation failed to submit, twice in a row',
        )
        spec.exit_code(
            405,
            'ERROR_SECOND_CONSECUTIVE_UNHANDLED_FAILURE',
            message='the calculation failed for an unknown reason, twice in a row',
        )
        spec.exit_code(
            500,
            'ERROR_MISSING_CRITICAL_OUTPUT',
            message='Missing critical output for inspecting the status of the calculation.',
        )
        spec.exit_code(
            501,
            'ERROR_OTHER_INTERVENTION_NEEDED',
            message='Cannot handle the error - inputs are likely need to be revised manually. Message: {message}',
        )
        spec.exit_code(
            502,
            'ERROR_CALCULATION_NOT_FINISHED',
            message='Cannot handle the error - the last calculation did not reach the end of execution.',
        )
        spec.exit_code(
            503,
            'ERROR_ELECTRONIC_STRUCTURE_NOT_CONVERGED',
            message='Cannot handle the error - the last calculation did not reach electronic convergence.',
        )
        spec.exit_code(
            504,
            'ERROR_IONIC_RELAXATION_NOT_CONVERGED',
            message='The ionic relaxation is not converged.',
        )
        spec.exit_code(
            505,
            'ERROR_UNCONVERGED_ELECTRONIC_STRUCTURE_IN_RELAX',
            message='At least one of the ionic steps during the relaxation has did not have converged '
            'electronic structure.',
        )
        spec.exit_code(
            700,
            'ERROR_NO_POTENTIAL_FAMILY_NAME',
            message='the user did not supply a potential family name',
        )
        spec.exit_code(
            701,
            'ERROR_POTENTIAL_VALUE_ERROR',
            message='ValueError was returned from get_potcars_from_structure',
        )
        spec.exit_code(
            702,
            'ERROR_POTENTIAL_DO_NOT_EXIST',
            message='the potential does not exist',
        )
        spec.exit_code(
            703,
            'ERROR_IN_PARAMETER_MASSAGER',
            message='the exception: {exception} was thrown while massaging the parameters',
        )

    def setup(self) -> None:
        super().setup()
        self.ctx.restart_calc = None
        self.ctx.vasp_did_not_execute = False
        self.ctx.last_calc_was_unfinished = False
        self.ctx.use_wavecar = True
        self.ctx.ignore_transient_nelm_breach = False  # Flag for ignoring the NELM breach during the relaxation
        self.ctx.verbose = None
        self.ctx.last_calc_remote_objects = []
        self.ctx.handler = AttributeDict()
        self.ctx.handler.nbands_increase_tries = 0

    def _init_parameters(self) -> Dict:
        """Collect input to the workchain in the converge namespace and put that into the parameters."""

        # At some point we will replace this with possibly input checking using the PortNamespace on
        # a dict parameter type. As such we remove the workchain input parameters as node entities. Much of
        # the following is just a workaround until that is in place in AiiDA core.
        parameters = inherit_and_merge_parameters(self.inputs)

        return parameters

    def verbose_report(self, *args, **kwargs) -> None:
        """Send report if self.ctx.verbose is True"""
        if self.ctx.verbose is True:
            self.report(*args, **kwargs)

    def prepare_inputs(self) -> None:
        """
        Enforce some settings for the restart folder and set parameters tags for a restart.
        This is called because launching the sub process.

        NOTE: This method should probably be refactored to give more control on what kind
        of restart is needed
        """
        # Check first if the calling workchain wants a restart in the same folder
        if 'restart_folder' in self.inputs:
            self.ctx.inputs.restart_folder = self.inputs.restart_folder

        # Then check if the workchain wants a restart
        if self.ctx.restart_calc and isinstance(self.ctx.restart_calc.process_class, self._process_class):
            self.ctx.inputs.restart_folder = self.ctx.restart_calc.outputs.remote_folder
            old_parameters = AttributeDict(self.ctx.inputs.parameters).copy()
            parameters = old_parameters.copy()
            # Make sure ISTART and ICHARG is set to read the relevant objects - if they exists
            if 'istart' in parameters and 'WAVECAR' in self.ctx.last_calc_remote_objects:
                # Correct in case of istart = 0
                if parameters.istart == 0 and self.ctx.use_wavecar:
                    parameters.istart = 1
            # Not using the WAVECAR - we make sure ISTART is 0
            if not self.ctx.use_wavecar:
                parameters.istart = 0
            if 'icharg' in parameters and 'CHGCAR' in self.ctx.last_calc_remote_objects:
                parameters.icharg = 1
            if parameters != old_parameters:
                self.ctx.inputs.parameters = parameters
                self.report('Enforced ISTART=1 and ICHARG=1 for restarting the calculation.')

        # Reset the list of valid remote objects and the restart calculation
        self.ctx.last_calc_remote_objects = []
        self.ctx.restart_calc = None

    def update_magmom(self, node: Optional[CalcJobNode] = None) -> None:
        """
        Update magmom from site magnetization information if available

        :param node: Calculation node to be used, defaults to the last launched calculation.
        """
        if self.is_noncollinear:
            self.report('Automatic carrying on magmom for non-collinear magnetism calculation is not implemented.')
            return

        if node is None:
            node = self.ctx.children[-1]

        if 'site_magnetization' in node.outputs:
            try:
                self.ctx.inputs.parameters['magmom'] = site_magnetization_to_magmom(
                    node.outputs.site_magnetization.get_dict()
                )
            except ValueError:
                pass

    def init_inputs(self) -> Optional[ExitCode]:
        """Make sure all the required inputs are there and valid, create input dictionary for calculation."""

        #### START OF THE COPY FROM VASPWorkChain ####
        #  - the only change is that the section about kpoints is deleted
        self.ctx.inputs = AttributeDict()
        self.ctx.inputs.parameters = self._init_parameters()
        # Set the code
        self.ctx.inputs.code = self.inputs.code

        # Set the structure (poscar)
        self.ctx.inputs.structure = self.inputs.structure

        # Set the kpoints (kpoints) - No longer needed, using kpoints/kpoints spacing as from below
        # self.ctx.inputs.kpoints = self.inputs.kpoints

        # Set settings
        unsupported_parameters = None
        skip_parameters_validation = False
        settings_dict = {}
        if self.inputs.get('settings'):
            self.ctx.inputs.settings = self.inputs.settings
            # Also check if the user supplied additional tags that is not in the supported file.
            settings_dict = self.ctx.inputs.settings.get_dict()
            unsupported_parameters = settings_dict.get('unsupported_parameters', unsupported_parameters)
            skip_parameters_validation = settings_dict.get('skip_parameters_validation', skip_parameters_validation)

        # Perform inputs massage to accommodate generalization in higher lying workchains
        # and set parameters.
        try:
            parameters_massager = ParametersMassage(
                self.ctx.inputs.parameters,
                unsupported_parameters,
                skip_parameters_validation=skip_parameters_validation,
            )
        except Exception as exception:  # pylint: disable=broad-except
            return self.exit_codes.ERROR_IN_PARAMETER_MASSAGER.format(exception=exception)  # pylint: disable=no-member
        try:
            # Only set if they exists
            # Set any INCAR tags
            self.ctx.inputs.parameters = parameters_massager.parameters.incar
            # Set any dynamics input (currently only for selective dynamics, e.g. custom write to POSCAR)
            self.ctx.inputs.dynamics = parameters_massager.parameters.dynamics
            # Here we could set additional override flags, but those are not relevant for this VASP plugin
        except AttributeError:
            pass

        # Set options
        # Options is very special, not storable and should be
        # wrapped in the metadata dictionary, which is also not storable
        # and should contain an entry for options
        if 'options' in self.inputs:
            options = {}
            options.update(self.inputs.options)
            self.ctx.inputs.metadata = {'options': options}
            # Override the parser name if it is supplied by the user.
            parser_name = self.ctx.inputs.metadata['options'].get('parser_name')
            if parser_name:
                self.ctx.inputs.metadata['options']['parser_name'] = parser_name
            # Set MPI to True, unless the user specifies otherwise
            withmpi = self.ctx.inputs.metadata['options'].get('withmpi', True)
            self.ctx.inputs.metadata['options']['withmpi'] = withmpi

        # Utilise default input/output selections
        self.ctx.inputs.metadata['options']['input_filename'] = 'INCAR'
        self.ctx.inputs.metadata['options']['output_filename'] = 'OUTCAR'

        # Make sure we also bring along any label and description set on the WorkChain to the CalcJob, it if does
        # not exists, set to empty string.
        if 'metadata' in self.inputs:
            label = self.inputs.metadata.get('label', '')
            description = self.inputs.metadata.get('description', '')
            if 'metadata' not in self.ctx.inputs:
                self.ctx.inputs.metadata = {}
            self.ctx.inputs.metadata['label'] = label
            self.ctx.inputs.metadata['description'] = description

        # Carry on site magnetization for initialization
        if 'site_magnetization' in self.inputs:
            magmom = site_magnetization_to_magmom(self.inputs.site_magnetization.get_dict())
            assert len(magmom) == len(self.inputs.structure.sites)
            self.ctx.inputs.parameters['magmom'] = magmom

        # Verify and set potentials (potcar)
        if not self.inputs.potential_family.value:
            self.report('An empty string for the potential family name was detected.')  # pylint: disable=not-callable
            return self.exit_codes.ERROR_NO_POTENTIAL_FAMILY_NAME  # pylint: disable=no-member
        try:
            self.ctx.inputs.potential = PotcarData.get_potcars_from_structure(
                structure=self.inputs.structure,
                family_name=self.inputs.potential_family.value,
                mapping=self.inputs.potential_mapping.get_dict(),
            )
        except ValueError as err:
            return compose_exit_code(self.exit_codes.ERROR_POTENTIAL_VALUE_ERROR.status, str(err))  # pylint: disable=no-member
        except NotExistent as err:
            return compose_exit_code(self.exit_codes.ERROR_POTENTIAL_DO_NOT_EXIST.status, str(err))  # pylint: disable=no-member

        # Store verbose parameter in ctx - otherwise it will not work after deserialization
        try:
            self.ctx.verbose = self.inputs.verbose.value
        except AttributeError:
            self.ctx.verbose = self._verbose

        # Set the charge density (chgcar)
        if 'chgcar' in self.inputs:
            self.ctx.inputs.charge_density = self.inputs.chgcar

        # Set the wave functions (wavecar)
        if 'wavecar' in self.inputs:
            self.ctx.inputs.wavefunctions = self.inputs.wavecar

        if 'vdw_kernel' in self.inputs:
            self.ctx.inputs.vdw_kernel = self.inputs.vdw_kernel

        ##### END OF THE COPY from VaspWorkChain   #####

        # Set the kpoints (kpoints)
        if 'kpoints' in self.inputs:
            self.ctx.inputs.kpoints = self.inputs.kpoints
        elif 'kpoints_spacing' in self.inputs:
            kpoints = KpointsData()
            kpoints.set_cell_from_structure(self.ctx.inputs.structure)
            kpoints.set_kpoints_mesh_from_density(self.inputs.kpoints_spacing.value * np.pi * 2)
            self.ctx.inputs.kpoints = kpoints
        else:
            raise InputValidationError("Must supply either 'kpoints' or 'kpoints_spacing'")

        # Setup LDAU keys
        if 'ldau_mapping' in self.inputs:
            ldau_settings = self.inputs.ldau_mapping.get_dict()
            ldau_keys = get_ldau_keys(self.ctx.inputs.structure, **ldau_settings)
            # Directly update the raw inputs passed to VaspCalculation
            self.ctx.inputs.parameters.update(ldau_keys)

        # Attach default monitors if not provided by the user
        if not self.inputs.get('monitors') and not settings_dict.get('no_default_monitors', False):
            self.ctx.inputs.monitors = {
                'stdout': Dict(dict={'entry_point': 'vasp.stdout'}),
                'loop_time': Dict(dict={'entry_point': 'vasp.loop_time', 'minimum_poll_interval': 600}),
            }
        return None

    def run_auto_parallel(self) -> bool:
        """Wether we should run auto-parallelisation test"""
        return 'auto_parallel' in self.inputs and self.inputs.auto_parallel.value is True

    def perform_autoparallel(self) -> None:
        """Dry run and obtain the best parallelisation settings"""

        self.report('Performing local dryrun for auto-parallelisation')  # pylint: disable=not-callable

        ind = prepare_process_inputs(self.ctx.inputs)

        nprocs = self.ctx.inputs.metadata['options']['resources']['tot_num_mpiprocs']

        # Take the settings pass it to the function
        kwargs = self.inputs.auto_parallel.get_dict()
        if 'cpus_per_node' not in kwargs:
            kwargs['cpus_per_node'] = self.inputs.code.computer.get_default_mpiprocs_per_machine()

        # If the dryrun errored, proceed the workchain
        try:
            scheme = get_jobscheme(ind, nprocs, **kwargs)
        except Exception as error:
            self.report(f'Dry-run errorred, process with cautions, message: {error.args}')  # pylint: disable=not-callable
            return

        if (scheme.ncore is None) or (scheme.kpar is None):
            self.report(f'Error NCORE: {scheme.ncore}, KPAR: {scheme.kpar}')  # pylint: disable=not-callable
            return

        parallel_opts = {'ncore': scheme.ncore, 'kpar': scheme.kpar}
        self.report(f'Found optimum KPAR={scheme.kpar}, NCORE={scheme.ncore}')  # pylint: disable=not-callable
        self.ctx.inputs.parameters.update(parallel_opts)
        self.out(
            'parallel_settings',
            Dict(dict={'ncore': scheme.ncore, 'kpar': scheme.kpar}).store(),
        )

    @property
    def is_noncollinear(self) -> bool:
        """Check if the calculation is a noncollinear one"""
        return self.ctx.inputs.parameters.get('lnoncollinear') or self.ctx.inputs.parameters.get('lsorbit')

    @override
    def on_except(self, exc_info: Tuple[Any, Exception, TracebackType]) -> None:
        """Handle excepted state."""
        try:
            last_calc = self.ctx.calculations[-1] if self.ctx.calculations else None
            if last_calc is not None:
                self.report(f'Last calculation: {last_calc!r}')  # pylint: disable=not-callable
                sched_err = last_calc.outputs.retrieved.get_file_content('_scheduler-stderr.txt')
                sched_out = last_calc.outputs.retrieved.get_file_content('_scheduler-stdout.txt')
                self.report(f'Scheduler output:\n{sched_out or ""}')  # pylint: disable=not-callable
                self.report(f'Scheduler stderr:\n{sched_err or ""}')  # pylint: disable=not-callable
        except AttributeError:
            self.report(
                'No calculation was found in the context. '  # pylint: disable=not-callable
                'Something really awful happened. '
                'Please inspect messages and act.'
            )

        return super().on_except(exc_info)

    @override
    def on_terminated(self) -> None:
        """
        Clean the working directories of all child calculation jobs if `clean_workdir=True` in the inputs and
        the calculation is finished without problem.
        """
        # Directly called the WorkChain method as this method replaces that of the BaseRestartWorkChain
        WorkChain.on_terminated(self)

        if self.inputs.clean_workdir.value is False:  # type: ignore[union-attr]
            self.report('remote folders will not be cleaned')
            return

        if not self.ctx.is_finished:  # type: ignore[union-attr]
            self.report('remote folders will not be cleaned because the workchain finished with error.')
            return

        # Find the remote folder of the last calculation which should be kept from cleaning
        out_remote_pk = None
        if self.inputs.keep_last_workdir.value is True:
            out_remote_pk = self.outputs.remote_folder.pk

        cleaned_calcs = []
        for called_descendant in self.node.called_descendants:
            if isinstance(called_descendant, CalcJobNode):
                try:
                    remote_folder = called_descendant.outputs.remote_folder
                    if remote_folder.pk != out_remote_pk:
                        remote_folder._clean()  # pylint: disable=protected-access
                    cleaned_calcs.append(str(called_descendant.pk))
                except (IOError, OSError, KeyError):
                    pass

        if cleaned_calcs:
            self.report(f'cleaned remote folders of calculations: {" ".join(cleaned_calcs)}')

    @process_handler(priority=2000, enabled=False)
    def handler_always_attach_outputs(self, node: CalcJobNode) -> Optional[ProcessHandlerReport]:
        """
        Handle the case where we attach the outputs even if underlying child calculation ends up
        with some exit status.
        """

        # Only active this error handler at the last iteration
        if node.is_finished_ok or self.ctx.iteration < self.inputs.max_iterations.value:
            return None

        # Attach all outputs from the last workchain
        self.report('At the last iteration - attaching outputs from the last workchain.')
        self.report('WARNING: The attached outputs may contain incorrect results - proceed with caution.')

        # Attach the required outputs defined in the spec
        for name, port in self.spec().outputs.items():
            try:
                output = node.get_outgoing(link_label_filter=name).one().node
            except ValueError:
                if port.required:
                    self.report(f"required output '{name}' was not an output of {self.ctx.process_name}<{node.pk}>")
            else:
                self.out(name, output)

        # Try to get some meaningful exit codes using the sanity check handler
        # Always generate a handler report with do_break so no more handlers will be run and
        # overwrite the error code.
        report = self._calculation_sanity_checks(node)
        if report:
            self.report('Problems during checks of the outputs. The corresponding `exit_code` will be returned.')
            return ProcessHandlerReport(exit_code=report.exit_code, do_break=True)
        return ProcessHandlerReport(exit_code=self.exit_codes.ERROR_MAXIMUM_ITERATION_EXCEEDED, do_break=True)

    @process_handler(priority=1100, exit_codes=VaspCalculation.exit_codes.ERROR_VASP_DID_NOT_EXECUTE)
    def handler_calculation_did_not_run(self, node: CalcJobNode) -> Optional[ProcessHandlerReport]:
        """Handle the case where the calculation is not performed"""
        if self.ctx.vasp_did_not_execute:
            self.report(f'{node} did not execute, and this is the second time - aborting.')
            return ProcessHandlerReport(
                do_break=True,
                exit_code=self.exit_codes.ERROR_OTHER_INTERVENTION_NEEDED.format(
                    message='VASP executable did not run on the remote computer.'
                ),
            )

        self.report(f'{node} did not execute - try again')
        self.ctx.vasp_did_not_execute = True
        return ProcessHandlerReport(do_break=True)

    @process_handler(priority=1000)
    def handler_misc_not_exist(self, node: CalcJobNode) -> Optional[ProcessHandlerReport]:
        """
        Handle the case where misc output is not available, in which case we cannot do anything for it.
        """
        # Check if the run is converged electronically
        if 'misc' not in node.outputs:
            self.report('Cannot found `misc` outputs - please check the process reports for issues.')
            return ProcessHandlerReport(exit_code=self.exit_codes.ERROR_MISSING_CRITICAL_OUTPUT, do_break=True)  # pylint: disable=no-member
        return None

    @process_handler(priority=910, exit_codes=[VaspCalculation.exit_codes.ERROR_DID_NOT_FINISH])
    def handler_unfinished_calc_ionic(self, node: CalcJobNode) -> Optional[ProcessHandlerReport]:
        """
        Handled the problem such that the calculation is not finished, e.g. did not reach the
        end of execution.

        If WAVECAR exists, just resubmit the calculation with the restart folder.

        If it is a geometry optimization, attempt to restart with output structure + WAVECAR.
        """

        # Check it is a geometry optimization
        incar = self.ctx.inputs.parameters
        if incar.get('nsw', -1) > 0:
            if 'structure' not in node.outputs:
                self.report('Performing a geometry optimization but the output structure is not found.')
                return ProcessHandlerReport(
                    do_break=True,
                    exit_code=self.exit_codes.ERROR_OTHER_INTERVENTION_NEEDED.format(
                        message='No output structure for restart.'
                    ),
                )  # pylint: disable=no-member
            self.report('Continuing geometry optimization using the last geometry.')
            self.ctx.inputs.structure = node.outputs.structure
            self._setup_restart(node)
            self.update_magmom(node)
            return ProcessHandlerReport(do_break=True)
        return None

    @process_handler(
        priority=799,
        enabled=False,
        exit_codes=[VaspCalculation.exit_codes.ERROR_DID_NOT_FINISH],
    )
    def handler_unfinished_calc_ionic_alt(self, node: CalcJobNode) -> Optional[ProcessHandlerReport]:
        """
        Handled the problem such that the calculation is not finished, e.g. did not reach the
        end of execution.

        If WAVECAR exists, just resubmit the calculation with the restart folder.

        If it is a geometry optimization, attempt to restart with output structure + WAVECAR.
        """

        # Check it is a geometry optimization
        incar = self.ctx.inputs.parameters
        if incar.get('nsw', -1) > 0:
            if 'structure' not in node.outputs:
                self.report('Performing a geometry optimization but the output structure is not found.')
                return ProcessHandlerReport(
                    do_break=True,
                    exit_code=self.exit_codes.ERROR_OTHER_INTERVENTION_NEEDED.format(
                        message='No output structure for restart.'
                    ),
                )  # pylint: disable=no-member
            self.report('Continuing geometry optimization using the last geometry.')
            self.ctx.inputs.structure = node.outputs.structure
            self._setup_restart(node)
            self.update_magmom(node)
            return ProcessHandlerReport(do_break=True)
        return None

    @process_handler(priority=798, enabled=False)
    def handler_unfinished_calc_generic_alt(self, node: CalcJobNode) -> Optional[ProcessHandlerReport]:
        """
        A generic handler for unfinished calculations, we attempt to restart it once.
        """

        # Only act on this specific return code, otherwise we reset the flag
        if node.exit_status != VaspCalculation.exit_codes.ERROR_DID_NOT_FINISH.status:
            self.ctx.last_calc_was_unfinished = False
            return None

        if self.ctx.last_calc_was_unfinished:
            msg = (
                'The last calculation was not completed for the second time, potentially due to insufficient '
                'walltime/node failure. Please revise the resources request and/or input parameters.'
            )
            return ProcessHandlerReport(
                do_break=True,
                exit_code=self.exit_codes.ERROR_OTHER_INTERVENTION_NEEDED.format(message=msg),
            )  # pylint: disable=no-member
        self.report(
            (
                'The last calculation was not finished - restart using the same set of inputs. '
                'If it was due to transient problem this may fix it, fingers crossed.'
            )
        )
        self.ctx.last_calc_was_unfinished = True
        return ProcessHandlerReport(do_break=True)

    @process_handler(priority=900)
    def handler_unfinished_calc_generic(self, node: CalcJobNode) -> Optional[ProcessHandlerReport]:
        """
        A generic handler for unfinished calculations, we attempt to restart it once.
        """

        # Only act on this specific return code, otherwise we reset the flag
        if node.exit_status != VaspCalculation.exit_codes.ERROR_DID_NOT_FINISH.status:
            self.ctx.last_calc_was_unfinished = False
            return None

        if self.ctx.last_calc_was_unfinished:
            msg = (
                'The last calculation was not completed for the second time, potentially due to insufficient'
                'walltime/node failure. Please revise the resources request and/or input parameters.'
            )
            return ProcessHandlerReport(
                do_break=True,
                exit_code=self.exit_codes.ERROR_OTHER_INTERVENTION_NEEDED.format(message=msg),
            )  # pylint: disable=no-member
        self.report(
            (
                'The last calculation was not finished - restart using the same set of inputs. '
                'If it was due to transient problem this may fix it, fingers crossed.'
            )
        )
        self.ctx.last_calc_was_unfinished = True
        return ProcessHandlerReport(do_break=True)

    @process_handler(
        priority=800,
        enabled=False,
        exit_codes=[
            VaspCalculation.exit_codes.ERROR_ELECTRONIC_NOT_CONVERGED,
            VaspCalculation.exit_codes.ERROR_IONIC_NOT_CONVERGED,
            VaspCalculation.exit_codes.ERROR_DID_NOT_FINISH,
            VaspCalculation.exit_codes.ERROR_VASP_CRITICAL_ERROR,
            VaspCalculation.exit_codes.ERROR_OVERFLOW_IN_XML,
        ],
    )
    def handler_electronic_conv_alt(self, node: CalcJobNode) -> Optional[ProcessHandlerReport]:  # pylint: disable=too-many-return-statements,too-many-branches
        """Handle electronic convergence problem"""
        incar = node.inputs.parameters.get_dict()
        run_status = node.outputs.misc['run_status']
        notifications = node.outputs.misc['notifications']
        nelm = run_status['nelm']
        algo = incar.get('algo', 'normal')

        # In case of ionic convergence problem, we also act if electronic convergence problem has been reported.
        if node.exit_status in [
            VaspCalculation.exit_codes.ERROR_IONIC_NOT_CONVERGED.status,
            VaspCalculation.exit_codes.ERROR_DID_NOT_FINISH,
        ]:
            perform_fix = False
            if run_status['consistent_nelm_breach']:
                self.report(
                    'The NELM limit has been breached in all ionic steps - proceed to take actions'
                    'for improving convergence.'
                )
                perform_fix = True
            elif run_status['contains_nelm_breach']:
                # Then there are some breaches in the ionic cycles
                if self.ctx.ignore_transient_nelm_breach:
                    self.report(
                        'WARNING: NELM limit breached in some ionic steps but requested to ignore this -'
                        'no action taken.'
                    )
                    perform_fix = False
                else:
                    self.report(
                        'The NELM limit has been breached in some ionic steps - proceed to take actions for'
                        ' improving convergence.'
                    )
                    perform_fix = True
            if not perform_fix:
                return None

        if node.exit_status == VaspCalculation.exit_codes.ERROR_VASP_CRITICAL_ERROR.status:
            # Make sure we only continue in this handler for a selected set of the critical errors.
            if not any(item in node.exit_message for item in ['EDDRMM', 'EDDDAV', 'The topmost band is occupied']):
                # We have some other critical error not to handle here
                return None

        # Check if we need to add more bands
        for item in notifications:
            if item['name'] == 'bandocc' and self.ctx.handler.nbands_increase_tries < 5:
                try:
                    nbands = run_status['nbands']
                    # Increase nbands with 10%
                    nbands_new = math.ceil(nbands * 1.1)
                    self.report(f'Changing NBANDS from {nbands} to {nbands_new}')
                    self.ctx.handler.nbands_increase_tries += 1
                    incar['nbands'] = nbands_new
                    self.ctx.inputs.parameters.update(incar)
                    return ProcessHandlerReport(do_break=True)
                except KeyError:
                    self.report(
                        'The topmost band is occupied but did not locate the nbands entry in run_status, '
                        'so no way to do corrections.'
                    )
                    return ProcessHandlerReport(
                        exit_code=self.exit_codes.ERROR_MISSING_CRITICAL_OUTPUT,
                        do_break=True,
                    )  # pylint: disable=no-member

        if nelm < 300:
            # Standard NELM might be a bit low, so increase a bit
            incar['nelm'] = 300
            # Here we can just continue from previous run
            self._setup_restart(node)
            self.ctx.inputs.parameters.update(incar)
            self.report(f'Changing NELM from {nelm} to 300.')
            return ProcessHandlerReport(do_break=True)

        # Let us start or continue to switch algorithms and reduce try list
        try:
            if self.ctx.handler.get('remaining_algos', None) is None:
                self.ctx.handler.remaining_algos = self._algo_switching[algo.lower()]
            new_algo = self.ctx.handler.remaining_algos.pop(0)
            incar['algo'] = new_algo
            self.ctx.inputs.parameters.update(incar)
            self.report(f'Changing ALGO from {algo.lower()} to {new_algo}')
            return ProcessHandlerReport(do_break=True)
        except IndexError:
            self.report('No more algorithms to try and we still have not reached electronic convergence.')

        self.report('No additional fixes can be applied to improve the electronic convergence - aborting.')
        return ProcessHandlerReport(
            do_break=True,
            exit_code=self.exit_codes.ERROR_OTHER_INTERVENTION_NEEDED.format(
                message='Cannot apply fix for reaching electronic convergence.'
            ),
        )

    @process_handler(
        priority=800,
        exit_codes=[
            VaspCalculation.exit_codes.ERROR_ELECTRONIC_NOT_CONVERGED,
            VaspCalculation.exit_codes.ERROR_IONIC_NOT_CONVERGED,
            VaspCalculation.exit_codes.ERROR_DID_NOT_FINISH,
        ],
    )
    def handler_electronic_conv(self, node: CalcJobNode) -> Optional[ProcessHandlerReport]:
        """Handle electronic convergence problem"""
        incar = node.inputs.parameters.get_dict()
        run_status = node.outputs.misc['run_status']
        nelm = run_status['nelm']
        algo = incar.get('algo', 'normal')

        # In case of ionic convergence problem, we also act if electronic convergence problem has been reported.
        if node.exit_status in [
            VaspCalculation.exit_codes.ERROR_IONIC_NOT_CONVERGED.status,
            VaspCalculation.exit_codes.ERROR_DID_NOT_FINISH,
        ]:
            perform_fix = False
            if run_status['consistent_nelm_breach']:
                self.report(
                    'The NELM limit has been breached in all ionic steps - proceed to take'
                    ' actions for improving convergence.'
                )
                perform_fix = True
            elif run_status['contains_nelm_breach']:
                # Then there are some breaches in the ionic cycles
                if self.ctx.ignore_transient_nelm_breach:
                    self.report(
                        'WARNING: NELM limit breached in some ionic steps but requested to ignore this'
                        ' - no action taken.'
                    )
                    perform_fix = False
                else:
                    self.report(
                        'The NELM limit has been breached in some ionic steps - proceed to take actions'
                        ' for improving convergence.'
                    )
                    perform_fix = True
            if not perform_fix:
                return None

        if algo.lower() in ('fast', 'veryfast'):
            incar['algo'] = 'normal'
            self._setup_restart(node)
            self.ctx.inputs.parameters.update(incar)
            self.report(f'Setting ALGO=normal from ALGO={algo.lower()}')
            return ProcessHandlerReport(do_break=True)

        # The logic below only works for algo=normal
        if algo.lower() == 'normal':
            # First try - Increase NELM
            if nelm < 150:
                incar['nelm'] = 150
                self._setup_restart(node)
                self.ctx.inputs.parameters.update(incar)
                self.report('Setting NELM to 150')
                return ProcessHandlerReport(do_break=True)
            # Adjust AMIX value if NELM is already high
            amix = incar.get('amix', 0.4)
            amix_steps = [0.2, 0.1, 0.05]
            for amix_target in amix_steps:
                if amix > amix_target:
                    incar['amix'] = amix_target
                    # Increase NELM in the mean time - smaller amplitude requires more cycles but more stable.
                    incar['nelm'] = nelm + 20
                    self._setup_restart(node)
                    self.ctx.inputs.parameters.update(incar)
                    self.report(f'Reducing AMIX to {incar["amix"]}')
                    return ProcessHandlerReport(do_break=True)
            # Change to ALGO if options have been exhausted
            incar['algo'] = 'all'
            self.ctx.inputs.parameters.update(incar)
            self._setup_restart(node)
            self.report('Switching to ALGO = ALL')
            return ProcessHandlerReport(do_break=True)
        self.report('No additional fixes can be applied to improve the electronic convergence - aborting.')
        return ProcessHandlerReport(
            do_break=True,
            exit_code=self.exit_codes.ERROR_OTHER_INTERVENTION_NEEDED.format(
                message='Cannot apply fix for reaching electronic convergence.'
            ),
        )

    @process_handler(
        priority=510,
        exit_codes=[VaspCalculation.exit_codes.ERROR_IONIC_NOT_CONVERGED],
        enabled=False,
    )
    def handler_ionic_conv_enhanced(self, node: CalcJobNode) -> Optional[ProcessHandlerReport]:  # pylint: disable=too-many-return-statements, too-many-branches
        """
        Enhanced handling of ionic relaxation problem beyond simple restarts.

        This is only used when the calculation is having difficulties reaching the
        convergence. This handler should be applied before the standard handler which
        breaks the handling cycle.
        """

        if 'structure' not in node.outputs:
            self.report('Performing a geometry optimization but the output structure is not found.')
            return ProcessHandlerReport(
                do_break=True,
                exit_code=self.exit_codes.ERROR_OTHER_INTERVENTION_NEEDED.format(
                    message='No output structure for restarting ionic relaxation.'
                ),
            )  # pylint: disable=no-member

        # The simplest solution - resubmit the calculation again
        child_nodes = self.ctx.children
        child_miscs = [node.outputs.misc for node in child_nodes]

        self.update_magmom(node)

        # Enhanced handler only takes place after 3 trials
        if len(child_miscs) < 3:
            return None

        natom = len(self.ctx.inputs.structure.sites)
        # Number of iterations
        ionic_iterations = [misc['run_status']['last_iteration_index'][0] for misc in child_miscs]

        # Output energies
        energies = []
        for misc in child_miscs:
            energies.append(misc.get('total_energies', {}).get('energy_extrapolated'))
        if all([eng is not None for eng in energies[-3:]]):  # pylint: disable=use-a-generator
            de_per_atom = np.diff(energies) / natom
        else:
            return None

        # First check if dE is very small
        if np.all(np.abs(de_per_atom) < 1e-5):
            msg = (
                'The total energy difference between the last two step is smaller than 1e-5 /atom'
                '- please consider to revise the cutoff value of the ionic steps.'
            )
            self.report(msg)
            return ProcessHandlerReport(
                do_break=True,
                exit_code=self.exit_codes.ERROR_OTHER_INTERVENTION_NEEDED.format(msg),
            )

        # Check if there are very few step performed per launch. Because VASP does not carry over
        # the internal parameters of the optimizer, this can make convergence slower.
        if ionic_iterations[-1] < 5:
            msg = (
                'Less than 5 iterations performed in the last launch - '
                'please consider submitting the jobs with revised resources request.'
            )
            self.report(msg)
            return ProcessHandlerReport(
                do_break=True,
                exit_code=self.exit_codes.ERROR_OTHER_INTERVENTION_NEEDED.format(msg),
            )

        # Warn about very unusually large number of steps and switch IBRION if needed.
        # Total degrees of freedom
        dof = 3 * natom
        isif = self.ctx.inputs.parameters.get('isif', 2)
        if isif == 3:
            dof += 6

        if sum(ionic_iterations) > dof + 10:
            self.report(f'Unusually large number of iterations performed for the degrees of freedom: {dof}')
            ibrion = self.ctx.inputs.parameters.get('ibrion')
            # In this case alternate between different relaxation algorithms
            if ibrion == 2:
                self.ctx.inputs.parameters['ibrion'] = 1
                self.ctx.inputs.parameters['potim'] = 0.3
                self.ctx.inputs.structure = node.outputs.structure
                self.report('Switching to IBRION=1 from IBRION=2 with POTIM = 0.3')
                return ProcessHandlerReport(do_break=True)
            if ibrion == 1:
                self.ctx.inputs.parameters['ibrion'] = 2
                self.ctx.inputs.parameters['potim'] = 0.1
                self.ctx.inputs.structure = node.outputs.structure
                self.report('Switching to IBRION=2 from IBRION=1 with POTIM = 0.1')
                return ProcessHandlerReport(do_break=True)

        # Check if energies are increasing without significant volume change
        vol_changes = []
        for child in child_nodes:
            inp_vol = child.inputs.structure.get_cell_volume()
            out_vol = child.outputs.structure.get_cell_volume()
            vol_changes.append(out_vol / inp_vol - 1.0)
        vol_changes = np.array(vol_changes)

        vol_tol = 0.03
        if np.all(de_per_atom > 0.0) and np.all(abs(vol_changes[-2:]) < vol_tol):
            msg = 'Energy increasing for the last two iterations - something can be very wrong...'
            self.report(msg)
            return ProcessHandlerReport(
                do_break=True,
                exit_code=self.exit_codes.ERROR_OTHER_INTERVENTION_NEEDED.format(msg),
            )

        self.report('No fixes can be applied for ionic convergence.')
        return None

    @process_handler(priority=505, exit_codes=[VaspCalculation.exit_codes.ERROR_IONIC_NOT_CONVERGED])
    def handler_ionic_conv(self, node: CalcJobNode) -> Optional[ProcessHandlerReport]:
        """Handle ionic convergence problem"""
        if 'structure' not in node.outputs:
            self.report('Performing a geometry optimization but the output structure is not found.')
            return ProcessHandlerReport(
                do_break=True,
                exit_code=self.exit_codes.ERROR_OTHER_INTERVENTION_NEEDED.format(
                    message='No output structure for restarting ionic relaxation.'
                ),
            )  # pylint: disable=no-member
        # The simplest solution - resubmit the calculation again
        self.report('Continuing geometry optimization using the last geometry.')
        self.ctx.inputs.structure = node.outputs.structure
        self._setup_restart(node)
        self.update_magmom(node)
        return ProcessHandlerReport(do_break=True)

    @process_handler(priority=400, exit_codes=[VaspCalculation.exit_codes.ERROR_VASP_CRITICAL_ERROR])
    def handler_vasp_critical_error(self, node: CalcJobNode) -> Optional[ProcessHandlerReport]:
        """
        Check if the calculation contain any critical error.
        """
        notification = node.outputs.misc['notifications']
        message = (
            f'Critical error detected in the notifications: {", ".join([item.get("name") for item in notification])}'
        )
        self.report(message + ' - aborting.')
        return ProcessHandlerReport(
            do_break=True,
            exit_code=self.exit_codes.ERROR_OTHER_INTERVENTION_NEEDED.format(message=message),
        )

    @process_handler(priority=5)
    def check_misc_output(self, node: CalcJobNode) -> Optional[ProcessHandlerReport]:
        """
        Check if misc output exists.
        """
        misc = node.outputs.misc.get_dict()
        if 'run_status' not in misc:
            self.report('`run_status` is not found in misc - cannot verify the integrity of the child calculation.')
            return ProcessHandlerReport(exit_code=self.exit_codes.ERROR_MISSING_CRITICAL_OUTPUT, do_break=True)
        return None

    @process_handler(priority=4)
    def check_calc_is_finished(self, node: CalcJobNode) -> Optional[ProcessHandlerReport]:
        """
        Check if the calculation has reached the end of execution.
        """
        misc = node.outputs.misc.get_dict()
        run_status = misc['run_status']
        if not run_status.get('finished'):
            self.report(f'The child calculation {node} did not reach the end of execution.')
            return ProcessHandlerReport(exit_code=self.exit_codes.ERROR_CALCULATION_NOT_FINISHED, do_break=True)
        return None

    @process_handler(priority=3)
    def check_electronic_converged(self, node: CalcJobNode) -> Optional[ProcessHandlerReport]:
        """
        Check if the calculation has converged electronic structure.
        """
        misc = node.outputs.misc.get_dict()
        run_status = misc['run_status']
        # Check that the electronic structure is converged
        if not run_status.get('electronic_converged'):
            self.report(f'The child calculation {node} does not possess a converged electronic structure.')
            return ProcessHandlerReport(
                exit_code=self.exit_codes.ERROR_ELECTRONIC_STRUCTURE_NOT_CONVERGED,
                do_break=True,
            )  # pylint: disable=no-member
        if run_status.get('contains_nelm_breach'):
            if self.ctx.ignore_transient_nelm_breach:
                self.report(
                    'The calculation contains at least one electronic minimization '
                    'that was truncated. It should thus not be considered converged. '
                    'Upon request from user, this is ignored.'
                )
            else:
                self.report(
                    'The calculation contains at least one electronic minimization '
                    'that is truncated. It should thus not be considered converged. '
                    'Treating the calculation as failed. Please inspect, maybe it is salvageable.'
                )
                return ProcessHandlerReport(
                    exit_code=self.exit_codes.ERROR_UNCONVERGED_ELECTRONIC_STRUCTURE_IN_RELAX,
                    do_break=True,
                )  # pylint: disable=no-member

        return None

    @process_handler(priority=2)
    def check_ionic_converged(self, node: CalcJobNode) -> Optional[ProcessHandlerReport]:
        """
        Check if the calculation has converged ionic structure.
        """

        # Check if we have requested to ignore ionic convergence check at calculation level
        # If so, then this handler should be by-passed
        if 'settings' in node.inputs:
            settings = node.inputs.settings.get_dict()
            if not settings.get('CHECK_IONIC_CONVERGENCE', True):
                return None

        misc = node.outputs.misc.get_dict()
        run_status = misc['run_status']

        # Check that the ionic structure is converged
        if run_status.get('ionic_converged') is False:
            self.report(f'The child calculation {node} did not have converged ionic structure.')
            return ProcessHandlerReport(
                exit_code=self.exit_codes.ERROR_IONIC_RELAXATION_NOT_CONVERGED,
                do_break=True,
            )  # pylint: disable=no-member
        return None

    # In this workchain we default to ignore the NELM breaches in the middle of the calculation
    @process_handler(priority=850, enabled=True)
    def ignore_nelm_breach_relax(self, node: CalcJobNode) -> None:
        """
        Not a actual handler but works as a switch to bypass checks for NELM breaches
         in the middle of an ionic relaxation.
        """
        _ = node
        self.ctx.ignore_transient_nelm_breach = True

    def _calculation_sanity_checks(self, node: CalcJobNode) -> Optional[ProcessHandlerReport]:  # pylint: disable=unused-argument
        """
        Perform additional sanity checks on successfully completed calculation.
        This method acts invokes the 'check' handlers to check the calculations and abort the workchain if any
        problem is found. This is useful when all of the corresponding error handlers are disabled, and allow
        one to avoid the default behaviour of restarting the calculation one more times regardlessly with
        unhandled errors.
        """
        checks = [
            self._check_misc_output,
            self._check_calc_is_finished,
            self._check_electronic_converged,
            self._check_ionic_converged,
        ]

        # Go though the checks one after another, return report if necessary
        last_report = None
        for check in checks:
            report = check(node)
            if report:
                if report.do_break:
                    return report
                last_report = report
        return last_report

    def _update_last_calc_objects(self, node: CalcJobNode) -> List[str]:
        """
        Connect to the remote and find the valid objects in th calculation folder

        Only update if the entry is empty in order to avoid too many connections to the remote.
        """
        if not self.ctx.last_calc_remote_objects:
            self.ctx.last_calc_remote_objects = list_valid_objects_in_remote(node.outputs.remote_folder)
        return self.ctx.last_calc_remote_objects

    def _setup_restart(self, node: CalcJobNode) -> bool:
        """
        Check the existence of any restart objects, if any of them eixsts use the last calculation
        for restart.
        """
        self._update_last_calc_objects(node)
        if 'WAVECAR' in self.ctx.last_calc_remote_objects or 'CHGCAR' in self.ctx.last_calc_remote_objects:
            self.ctx.restart_calc = node
            return True
        return False


def list_valid_objects_in_remote(remote: orm.RemoteData, path: str = '.', size_threshold: int = 0) -> List[str]:
    """
    List non-empty objects in the remote folder

    :param remote: The `RemoteFolder` node to be inspected.
    :param path: The relative path.
    :param size_threshold: The size threshold to treat the object as a valide one.

    :returns: A list of valid objects in the directory.
    """
    none_empty = []
    try:
        contents = remote.listdir_withattributes(path)
    except OSError:
        return []

    for obj in contents:
        if obj['attributes'].st_size > size_threshold and not obj['isdir']:
            none_empty.append(obj['name'])
    return none_empty


def potential_family_validator(family: orm.Str, _) -> None:
    """
    Validate the potential family input.

    :param faimly: The potential family to be validated.
    :raises ValueError: If the potential family is not valid.
    """
    if not family.value:
        raise InputValidationError('The potential family cannot be empty.')

    group = PotcarData.get_potcar_group(family.value)
    if group is None:
        raise InputValidationError(
            f'The potential family "{family.value}" is not found. '
            'Please use verdi data vasp.potcars tool to verify your settings.'
        )



"""
Some convenience mixins
"""

# ruff: noqa: PLC0415
from __future__ import annotations

from typing import Any


class WithBuilderUpdater:
    @classmethod
    def get_builder_updater(cls, *args: Any, **kwargs: Any) -> Any:
        """
        Return the corresponding builder updater class for the workchain.

        The arguments are passed directly to the underling `BuilderUpdater` constructor.
        """
        from .common.builder_updater import (
            VaspBandUpdater,
            VaspBuilderUpdater,
            VaspConvUpdater,
            VaspHybridBandUpdater,
            VaspMultiStageRelaxUpdater,
            VaspNEBUpdater,
            VaspRelaxUpdater,
        )

        if cls.__name__ == 'VaspWorkChain':
            return VaspBuilderUpdater(*args, **kwargs)
        elif cls.__name__ == 'VaspRelaxWorkChain':
            return VaspRelaxUpdater(*args, **kwargs)
        elif cls.__name__ == 'VaspMultiStageRelaxWorkChain':
            return VaspMultiStageRelaxUpdater(*args, **kwargs)
        elif cls.__name__ == 'VaspBandsWorkChain':
            return VaspBandUpdater(*args, **kwargs)
        elif cls.__name__ == 'VaspHybridBandsWorkChain':
            return VaspHybridBandUpdater(*args, **kwargs)
        elif cls.__name__ == 'VaspConvergenceWorkChain':
            return VaspConvUpdater(*args, **kwargs)
        elif cls.__name__ == 'VaspNEBWorkChain':
            return VaspNEBUpdater(*args, **kwargs)
        raise NotImplementedError('No builder updater found for workchain {}'.format(cls.__name__))



"""
Inheritance tools.

This sets up inheritance of the docstrings for use when we inherit
classes. This makes it simple to add further details to or replace the
docstring present on the base class.
"""

from __future__ import annotations

from typing import Callable, TypeVar

ClassType = TypeVar('ClassType')


def update_docstring(
    method_name: str, content: str, append: bool = True
) -> Callable[[type[ClassType]], type[ClassType]]:
    r"""
    Update docstring of (an inherited) class method.

    For subclasses that use hooks to change behavior of superclass methods.

    :param method_name: Name of the method to update
    :param content: Content to add or replace
    :param append: If true, append to the docstring, else overwrite it entirely.

    Example::

        class Base(object):
            def method(self, **kwargs):
                '''Print the base_arg kwarg.'''
                print kwargs.pop('base_arg')
                self.process_additional(**kwargs)

            def process_additional(self, **kwargs):
                pass


        @update_docstring('method', '\n\nAlso print the sub_arg kwarg.', append=True)
        class Subclass(Base):
            def process_additional(self, **kwargs):
                print kwargs['sub_arg']


        @update_docstring('method', 'Print all kwargs.', append=False)
        class Allprinter(Base):
            def process_additional(self, **kwargs):
                for _, value in kwargs.items():
                    print value
    """

    def wrapper(cls: type[ClassType]) -> type[ClassType]:
        """Update the method docstring and return the class."""
        if append:
            getattr(cls, method_name).__func__.__doc__ += content
        else:
            getattr(cls, method_name).__func__.__doc__ = content
        return cls

    return wrapper



"""
Module for easy access to aiida-vasp calculations.
"""

import os
import subprocess
import sys
import tempfile
from functools import wraps
from shutil import copyfileobj
from typing import Any, Callable

import click
from aiida import orm
from aiida.cmdline.commands.cmd_calcjob import calcjob_outputcat
from aiida.cmdline.params.arguments import (
    CALCULATION,
    PROCESS,
    WORKFLOW,
)
from aiida.cmdline.utils.echo import (
    echo_critical,
    echo_error,
    echo_info,
    echo_success,
)
from aiida.common.escaping import escape_for_bash
from aiida.common.exceptions import NotExistent
from aiida.orm import CalcJobNode, QueryBuilder, WorkChainNode

from aiida_vasp.utils.aiida_utils import cmp_load_verdi_data
from aiida_vasp.utils.export import export_vasp

VERDI_DATA = cmp_load_verdi_data()


@VERDI_DATA.group('vasp-tools')
def tools() -> None:
    """Tool for aiida-vasp related data"""


@tools.command('export')
@PROCESS('process')
@click.argument('folder')
@click.option(
    '--include-potcar',
    default=False,
    is_flag=True,
    help='Whether to include POTCAR in the export folder',
)
@click.option(
    '--decompress',
    default=False,
    is_flag=True,
    help='Wether to decompress the contents',
)
def export(process: orm.ProcessNode, folder: str, decompress: bool, include_potcar: bool) -> None:
    """Export a VASP calculation, works for both `VaspCalculation` or `VaspWorkChain`"""
    export_vasp(process, folder, decompress, include_potcar)


def select_calcjob_from_work(func: Callable) -> Callable:
    """Select calcjob from work"""

    @wraps(func)
    def wrapper(*args, **kwargs):
        calcjob = kwargs['calcjob']
        index = kwargs.pop('index', 0)

        if isinstance(calcjob, orm.WorkChainNode):
            valid = [process for process in calcjob.called_descendants if not process.is_finished]
            if len(valid) > 0:
                echo_info(
                    f'More than one calculations are still running: {", ".join(str(process.pk) for process in valid)}.'
                    f' Selected {valid[index].pk}'
                )
                calcjob = valid[index]
            else:
                calcjob = valid[0]
        kwargs['calcjob'] = calcjob
        kwargs['index'] = index

        func_ = click.option(
            '--index',
            '-i',
            help='The index of the calculation to cat if multiple calculations are still running',
            default=0,
        )(func)
        return func_(*args, **kwargs)

    return wrapper


@tools.command('remotecat')
@PROCESS('calcjob')
@click.argument('fname')
@click.option('--save-to', '-s', help='Name of the file to save to')
@select_calcjob_from_work
def remotecat(calcjob: orm.CalcJobNode, fname: str, save_to: str | None) -> None:
    """
    Print the conetent of a remote file to STDOUT

    This command for printing the content of a remote file to STDOUT.
    Useful for analysing running calculations.
    """

    rfolder = calcjob.outputs.remote_folder
    if save_to is None:
        fd, temppath = tempfile.mkstemp()
    else:
        temppath = save_to

    rfolder.getfile(fname, temppath)

    with open(temppath, 'rb') as fhandle:
        copyfileobj(fhandle, sys.stdout.buffer)

    if save_to is None:
        os.close(fd)
        os.remove(temppath)


@tools.command('remotepull')
@PROCESS('calcjob')
@click.argument('dest')
@click.option(
    '--max-size',
    '-m',
    help='Maximum size of the files to be retrieved - this is passed to rsync',
)
@select_calcjob_from_work
def remotepull(calcjob: orm.CalcJobNode, dest: str, max_size: str | None) -> None:
    """
    Pull a calculation folder from the remote

    This command for pull a calculation folder to a local folder.
    `rsync` is used for doing the heavy lifting.
    """

    rfolder = calcjob.outputs.remote_folder
    cmd_args = ['rsync', '-av']

    if max_size:
        cmd_args.extend(['--max-size', max_size])

    cmd_args.append(f'{rfolder.computer.hostname}:{rfolder.get_remote_path()}/')
    if not dest.endswith('/'):
        dest = dest + '/'
    cmd_args.append(dest)

    echo_info('Running commands: {}'.format(' '.join(cmd_args)))

    completed = subprocess.run(cmd_args, check=False)
    if completed.returncode != 0:
        echo_error('Failled to pull data using rsync')
    else:
        echo_success(f'Remote folder pulled to {dest}')


@tools.command('remotetail')
@CALCULATION('calcjob')
@click.argument('fname')
def remotetail(calcjob: orm.CalcJobNode, fname: str) -> None:
    """
    Follow a file on the remote computer

    This command will launch a ssh session dedicated for following a file
    using the `tail -f` command
    """

    try:
        transport = calcjob.get_transport()
    except NotExistent as exception:
        echo_critical(repr(exception))

    remote_workdir = calcjob.get_remote_workdir()

    if not remote_workdir:
        echo_critical('no remote work directory for this calcjob, maybe the daemon did not submit it yet')

    command = tailf_command(transport, remote_workdir, fname)
    os.system(command)


@tools.command('relaxcat')
@WORKFLOW('workflow')
@click.argument('fname')
def relaxcat(workflow: orm.WorkChainNode, fname: str) -> None:
    """Cat the output of the last calculation of a finished workflow"""

    q = QueryBuilder()
    q.append(WorkChainNode, filters={'id': workflow.id})
    q.append(WorkChainNode)
    q.append(CalcJobNode, tag='calc', project=['*', 'ctime'])
    q.order_by({'calc': {'ctime': 'desc'}})
    calc, _ = q.first()

    click.Context(calcjob_outputcat).invoke(calcjob_outputcat, calcjob=calc, path=fname)


def tailf_command(transport: Any, remotedir: str, fname: str) -> str:
    """
    Specific gotocomputer string to connect to a given remote computer via
    ssh and directly go to the calculation folder and then do tail -f of the target file.
    """

    further_params = []
    if 'username' in transport._connect_args:
        further_params.append('-l {}'.format(escape_for_bash(transport._connect_args['username'])))

    if transport._connect_args.get('port'):
        further_params.append('-p {}'.format(transport._connect_args['port']))

    if transport._connect_args.get('key_filename'):
        further_params.append('-i {}'.format(escape_for_bash(transport._connect_args['key_filename'])))

    further_params_str = ' '.join(further_params)

    connect_string = (
        """ "if [ -d {escaped_remotedir} ] ;"""
        """ then cd {escaped_remotedir} ; {bash_command} -c 'tail -f {fname}' ; else echo '  ** The directory' ; """
        """echo '  ** {remotedir}' ; echo '  ** seems to have been deleted, I logout...' ; fi" """.format(
            bash_command=transport._bash_command_str,
            escaped_remotedir=f"'{remotedir}'",
            remotedir=remotedir,
            fname=fname,
        )
    )

    cmd = 'ssh -t {machine} {further_params} {connect_string}'.format(
        further_params=further_params_str,
        machine=transport._machine,
        connect_string=connect_string,
    )
    return cmd



"""
The .win parser interface.

=========================

Contains routines to parse Wannier90 input files. Will in the future utilize
the parser in the Wannier90 plugin, but no input parser exists yet.
"""

import re
from typing import Any, Callable


class BaseKeyValueParser:  # pylint: disable=useless-object-inheritance
    """
    Common codebase for all parser utilities.

    This class provides utility methods for parsing key-value and line-based data.
    """

    empty_line = re.compile(r'[\r\n]\s*[\r\n]')

    @classmethod
    def line(cls, fobj_or_str: str | Any, d_type: type = str) -> Any:
        """
        Grab a line from a file object or string and convert it to ``d_type`` (default: ``str``).

        :param fobj_or_str: File object or string to read the line from.
        :type fobj_or_str: file or str
        :param d_type: Type to convert each item in the line to. Defaults to ``str``.
        :type d_type: type
        :return: Single value or list of values, depending on the number of items in the line.
        """
        if isinstance(fobj_or_str, str):
            line = fobj_or_str
        else:
            line = fobj_or_str.readline()
        # previously this was map instead of list comprehension
        res = [d_type(item) for item in line.split()]
        if len(res) == 1:
            return res[0]
        return res

    @classmethod
    def splitlines(cls, fobj_or_str: str | Any, d_type: type = float) -> list[Any]:
        """
        Split a chunk of text into a list of lines and convert each line to ``d_type`` (default: ``float``).

        :param fobj_or_str: File object or string to split into lines.
        :type fobj_or_str: file or str
        :param d_type: Type to convert each item in the line to. Defaults to ``float``.
        :type d_type: type
        :return: List of values, one per line.
        """
        if isinstance(fobj_or_str, str):
            lines = fobj_or_str.split('\n')
        else:
            lines = fobj_or_str.readlines()
        return [cls.line(line, d_type) for line in lines]


class KeyValueParser(BaseKeyValueParser):
    """
    Key and value parser.

    This base class provides utility functions for parsing files that are
    (mostly) in a ``key = value`` format.

    .. note::
        This class does not integrate with the ``VaspParser`` class currently.

    Example usage::

        import re
        from aiida_vasp.parsers.file_parsers.parser import KeyValueParser

        class ParamParser(KeyValueParser):
            def __init__(self, file_path):
                self._file_path = py.path.local(file_path)
                super().__init__()
                self.result = {}
            def convert_or_not(self, value):
                for converter in self.get_converter_iter():
                    converted = self.try_convert(value, converter)
                    if converted and 'value' in converted:
                        return converted['value']
                return value
            def parse_file(self):
                assignments = re.findall(self.assignment, self._file_path.read())
                return {key: self.convert_or_not(value)}

    Parses files like::

        StrParam = value_1
        FloatParam = 1.0
        BoolParam = True
    """

    assignment = re.compile(r'(\w+)\s*[=: ]\s*([^;\n]*);?')
    bool_true = re.compile(r'^T$')
    bool_false = re.compile(r'^F$')
    comment = True

    @classmethod
    def get_lines(cls, filename: str) -> list[str]:
        """
        Read all lines from a file.

        :param filename: Path to the file.
        :type filename: str
        :return: List of lines from the file.
        :rtype: list[str]
        """
        with open(filename, 'r', encoding='utf8') as input_file:
            lines = input_file.read().splitlines()
        return lines

    @classmethod
    def retval(cls, *args: Any, **kwargs: Any) -> dict[str, Any]:
        """
        Normalize return values from value conversion functions.

        :return: Dictionary with the value and any additional keyword arguments.
        :rtype: dict
        """
        val = list(args)
        if len(val) == 1:
            val = val[0]
        ret = {'value': val}
        ret.update(kwargs)
        return ret

    @classmethod
    def flatten(cls, lst: list[list[Any]]) -> list[Any]:
        """
        Flatten a list of lists into a single list.

        :param lst: List of lists.
        :type lst: list
        :return: Flattened list.
        :rtype: list
        """
        return [i for j in lst for i in j]

    @classmethod
    def find_kv(cls, line: str) -> list[tuple[str, str]]:
        """
        Find key-value pairs in a line using the assignment regex.

        :param line: Line to search for key-value pairs.
        :type line: str
        :return: List of (key, value) tuples.
        :rtype: list[tuple]
        """
        return re.findall(cls.assignment, line)

    @classmethod
    def float(cls, string_: str) -> dict[str, Any]:
        """
        Parse a string into a float value followed by a comment.

        :param string_: String to parse.
        :type string_: str
        :return: Dictionary with value and comment.
        :rtype: dict
        """
        vals = string_.split()
        value = float(vals.pop(0))
        comment = ' '.join(vals)
        return cls.retval(value, comment=comment)

    @classmethod
    def float_unit(cls, string_: str) -> dict[str, Any]:
        """
        Parse string into a float number with attached unit.

        :param string_: String to parse.
        :type string_: str
        :return: Dictionary with value, unit, and comment.
        :rtype: dict
        """
        vals = string_.split()
        value = float(vals.pop(0))
        unit = vals.pop(0) if vals else ''
        comment = ' '.join(vals)
        return cls.retval(value, unit, comment=comment)

    @classmethod
    def int(cls, string_: str) -> dict[str, Any]:
        """
        Parse a string into an integer value followed by a comment.

        :param string_: String to parse.
        :type string_: str
        :return: Dictionary with value and comment.
        :rtype: dict
        """
        vals = string_.split()
        value = int(vals.pop(0))
        comment = ' '.join(vals)
        return cls.retval(value, comment=comment)

    @classmethod
    def int_unit(cls, string_: str) -> dict[str, Any]:
        """
        Convert a string into a python value, associated unit and optional comment.

        :param string_: String to parse.
        :type string_: str
        :return: Dictionary with value, unit, and comment.
        :rtype: dict
        """
        vals = string_.split()
        value = int(vals.pop(0))
        unit = vals.pop(0) if vals else ''
        comment = ' '.join(vals)
        return cls.retval(value, unit, comment=comment)

    @classmethod
    def string(cls, string_: str) -> dict[str, Any]:
        """
        Parse a string into value and comment, assuming only the first word is the value.

        :param string_: String to parse.
        :type string_: str
        :return: Dictionary with value and comment.
        :rtype: dict
        """
        vals = string_.split()
        value = vals.pop(0)
        comment = ' '.join(vals)
        return cls.retval(value, comment=comment)

    @classmethod
    def bool(cls, string_: str) -> dict[str, Any]:
        """
        Parse string into a boolean value.

        :param string_: String to parse.
        :type string_: str
        :return: Dictionary with value and comment.
        :rtype: dict
        :raises ValueError: If the string does not match a boolean pattern.
        """
        vals = string_.split()
        bool_str = vals.pop(0)
        if re.match(cls.bool_true, bool_str):
            value = True
        elif re.match(cls.bool_false, bool_str):
            value = False
        else:
            raise ValueError(
                f'bool string {string_} did not match any of {[cls.bool_true.pattern, cls.bool_false.pattern]}'
            )
        comment = ' '.join(vals)
        return cls.retval(value, comment=comment)

    @classmethod
    def kv_list(cls, filename: str) -> list[Any]:
        """
        Read a file and return a list of key-value pairs for each line.

        :param filename: Path to the file.
        :type filename: str
        :return: List of key-value pairs.
        :rtype: list
        """
        with open(filename, 'r', encoding='utf8') as input_fo:
            kv_list = filter(None, map(cls.find_kv, input_fo))
        return kv_list

    @classmethod
    def kv_dict(cls, kv_list: list[Any]) -> dict[str, Any]:
        """
        Convert a list of key-value pairs into a dictionary.

        :param kv_list: List of key-value pairs.
        :type kv_list: list
        :return: Dictionary of key-value pairs.
        :rtype: dict
        """
        kv_dict = dict(cls.flatten(kv_list))
        return kv_dict

    @classmethod
    def clean_value(cls, str_value: str) -> dict[str, Any]:
        """
        Get the converted python value from a string.

        :param str_value: String value to convert.
        :type str_value: str
        :return: Dictionary with the converted value.
        :rtype: dict
        """
        if str_value == '':
            return cls.retval(str_value)
        cleaned_value = None
        converters = cls.get_converter_iter()
        while not cleaned_value:
            cleaned_value = cls.try_convert(str_value, converters.next())
        return cleaned_value

    @classmethod
    def get_converter_iter(cls) -> Any:
        """
        Get an iterator over the value converter functions in order.

        :return: Iterator over converter functions.
        """
        converter_order = [cls.bool, cls.int, cls.float, cls.string]
        return (i for i in converter_order)

    @classmethod
    def try_convert(cls, input_value: str, converter: Callable[[str], dict[str, Any]]) -> dict[str, Any] | None:
        """
        Try to convert the input string into a python value given a conversion function.

        :param input_value: Value to convert.
        :type input_value: str
        :param converter: Converter function to use.
        :type converter: callable
        :return: Dictionary with the converted value, or None if conversion failed.
        :rtype: dict or None
        """
        if not isinstance(input_value, str):
            return {'value': input_value}
        try:
            cleaned_value = converter(input_value)
        except ValueError:
            cleaned_value = {}

        if cleaned_value.get('value', None) is None:
            return None
        return cleaned_value


class WinParser(KeyValueParser):
    """
    Parses wannier90 input files.

    This parser extracts keywords, blocks, and comments from a Wannier90 ``.win`` input file.
    """

    block = re.compile(r'begin (?P<name>\w*)\s*\n\s*(?P<content>[\w\W]*)\s*\n\s*end \1')
    comment = re.compile(r'(!.*)\n?')

    def __init__(self, path: str) -> None:
        """
        Initialize the parser and parse the Wannier90 input file.

        :param path: Path to the Wannier90 .win file.
        :type path: str
        """
        super().__init__()
        self.result = {}
        with open(path, 'r', encoding='utf8') as winf:
            self.keywords, self.blocks, self.comments = WinParser.parse_win(winf)
        self.result.update(self.keywords)
        self.result.update(self.blocks)

    @classmethod
    def parse_win(cls, fobj_or_str: str | Any) -> tuple[dict[str, Any], dict[str, list[str]], list[str]]:
        """
        Parse a Wannier90 input file or string.

        :param fobj_or_str: File object or string containing the Wannier90 input.
        :type fobj_or_str: file or str
        :return: Tuple of (keywords dict, blocks dict, comments list).
        :rtype: tuple
        """
        if isinstance(fobj_or_str, str):
            content = fobj_or_str
        else:
            content = fobj_or_str.read()
        comments = re.findall(cls.comment, content)
        content = re.sub(cls.comment, '', content)
        blocks = re.findall(cls.block, content)
        content = re.sub(cls.block, '', content)
        kvd = dict(re.findall(cls.assignment, content))
        bld = {}
        for keyword, value in blocks:
            # do not split individual lines
            bld[keyword] = [line.strip() for line in value.split('\n')]
        return kvd, bld, comments
